/*------------------------------------------------------------------------------
 lib/djv_base/file_fnc.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_base/file_fnc.cxx"

#include "file.h"
#include <djv_base/math.h>
#include <FL/filename.H>
#include <algorithm>

namespace djv {
namespace file {

void seq_compress(List<File> * in)
{
  //DJV_DBG_OBJ2(_DBG, "seq_compress");
  //DJV_DBG_PRN("size = " << in->size());
  
  List<File> & list = *in;
  int cache = -1;
  uint i = 0;
  for (uint j = 0; j < list.size(); ++j) {
    //DJV_DBG_PRN("item = " << list[j]);
    
    const bool seq = list[j].seq_valid();
    if (seq) {
      
      if (cache != -1)
        if (list[cache].seq_add(list[j])) continue;
      
      cache = -1;
      for (uint k = 0; k < i; ++k) {
        if (SEQ == list[k].type)
          if (list[k].seq_add(list[j])) {
            cache = k;
            continue;
          }
      }
    }
    
    if (! seq || -1 == cache) {
      list[i] = list[j];
      if (seq) list[i].type = SEQ;
      ++i;
    }
  }
  //DJV_DBG_PRN("size = " << i);
  list.resize(i);
}

namespace {
inline bool relative_path(const Str & in)
{
  if ("." == in) ;
  else if (".." == in) ;
  else if (in.size() >= 2 && '.' == in[0] && ('/' == in[1] || '\\' == in[1])) ;
  else if (in.size() >= 3 && '.' == in[0] && '.' == in[1] &&
    ('/' == in[2] || '\\' == in[2])) ;
  else return false;
  return true;
}
}

void filter(List<File> * in, int filter, const List<Str> & glob_list)
{
  //DJV_DBG_OBJ2(_DBG, "filter");
  //DJV_DBG_PRN("filter = " << filter);
  //DJV_DBG_PRN("globs = " << glob_list);
  
  List<File> & list = *in;
  
  uint i = 0;
  for (uint j = 0; j < list.size(); ++j) {
    const Str name = list[j].get(-1, false);
    //DJV_DBG_PRN("name = " << name);
    
    bool valid = true;
    
    if ((filter & FILTER_FILE) && (FILE == list[j].type || SEQ == list[j].type))
      valid = false;
    
    if ((filter & FILTER_DIR) && DIR == list[j].type)
      valid = false;
    
    if ((filter & FILTER_HIDDEN) && '.' == name[0] && ! relative_path(name))
      valid = false;
    
    if (glob_list.size()) {
      uint k = 0;
      for (; k < glob_list.size(); ++k)
        if (fl_filename_match(name.c_str(), glob_list[k].c_str())) break;
      if (glob_list.size() == k)
        valid = false;
    }
    
    //DJV_DBG_PRN("valid = " << valid);
    if (valid) {
      if (i != j) list[i] = list[j];
      ++i;
    }
  }
  
  //DJV_DBG_PRN("size = " << i);
  list.resize(i);
}

namespace {

typedef bool (Compare)(const File &, const File &);

inline bool name(const File & a, const File & b) { return a.get() < b.get(); }
inline bool type(const File & a, const File & b) { return a.type < b.type; }
inline bool size(const File & a, const File & b) { return a.size < b.size; }
inline bool user(const File & a, const File & b) { return a.user < b.user; }
inline bool perm(const File & a, const File & b) { return a.perm < b.perm; }
inline bool time(const File & a, const File & b) { return a.time < b.time; }

inline bool name_r(const File & a, const File & b) { return name(b, a); }
inline bool type_r(const File & a, const File & b) { return type(b, a); }
inline bool size_r(const File & a, const File & b) { return size(b, a); }
inline bool user_r(const File & a, const File & b) { return user(b, a); }
inline bool perm_r(const File & a, const File & b) { return perm(b, a); }
inline bool time_r(const File & a, const File & b) { return time(b, a); }

}

void sort(List<File> * in, SORT sort, bool reverse)
{
  DJV_DBG_OBJ2(_DBG, "sort");
  DJV_DBG_PRN("in = " << in->size());
  DJV_DBG_PRN("sort = " << sort);
  DJV_DBG_PRN("reverse = " << reverse);
  
  Compare * compare = 0;
  switch (sort) {
    case SORT_NAME: compare = reverse ? name_r : name; break;
    case SORT_TYPE: compare = reverse ? type_r : type; break;
    case SORT_SIZE: compare = reverse ? size_r : size; break;
    case SORT_USER: compare = reverse ? user_r : user; break;
    case SORT_PERM: compare = reverse ? perm_r : perm; break;
    case SORT_TIME: compare = reverse ? time_r : time; break;
  }
  
  std::sort(in->begin(), in->end(), compare);
}

List<File> sort_dirs(const List<File> & in)
{
  DJV_DBG_OBJ2(_DBG, "sort_dirs");
  
  List<File> out, tmp;
  
  for (List<File>::const_iterator i = in.begin(); i != in.end(); ++i)
    if (DIR == i->type) out += *i;
    else tmp += *i;
  
  for (List<File>::const_iterator i = tmp.begin(); i != tmp.end(); ++i)
    out += *i;
  
  return out;
}

Str path_absolute(const Str & in)
{
  DJV_DBG_OBJ2(_DBG, "path_absolute");
  DJV_DBG_PRN("in = " << in);
  
  char tmp [str::cstr_size] = "";
  fl_filename_absolute(tmp, str::cstr_size, in.c_str());

  Str out(tmp);
  DJV_DBG_PRN("out = " << out);
  return out;
}

bool is_path_absolute(const Str & in)
{
  Str::size_type i = 0, j = in.find(':', 0);
  if (j != Str::npos) i = j + 1;
  return i < in.size() ? ('/' == in[i] || '\\' == in[i]) : false;
}

Str path_fix(const Str & in)
{
  //DJV_DBG_OBJ2(_DBG, "path_fix");
  //DJV_DBG_PRN("in = " << in);

  char tmp [str::cstr_size], path [str::cstr_size];
  fl_filename_expand(tmp, str::cstr_size, in.c_str());
  fl_filename_absolute(path, str::cstr_size, tmp);
  
  List<Str> list = str::split(path, "/\\");
  Str drive;
  if (list.size())
    if (list[0].find(':', 0) != Str::npos) {
      drive = list[0];
      list.erase(list.begin());
    }
  
  List<Str> list_tmp;
  for (List<Str>::const_iterator i = list.begin(); i < list.end(); ++i)
    if (".." == *i && list_tmp.size()) {
      if (list_tmp[list_tmp.size() - 1] != "..")
        list_tmp.erase(list_tmp.end() - 1);
      else
        list_tmp += *i;
    } else if ("." == *i && list_tmp.size()) {
    } else {
        list_tmp += *i;
    }
  
  const Str out = drive + path_separator +
    str::join(list_tmp, path_separator) +
    (list_tmp.size() ? path_separator : "");
  //DJV_DBG_PRN("out = " << out);
  return out;
}

Str path_posix(const Str & in)
{
  Str out = in;
  for (uint i = 0; i < out.size(); ++i)
    if (path_separator_list[1] == out[i])
      out[i] = path_separator_list[0];
  return out;
}

Str seq_to_str(const Seq & in)
{
  //DJV_DBG_OBJ2(_DBG, "seq_to_str");
  //DJV_DBG_PRN("in = " << in.start << "-" << in.end << ":" << in.pad);
  
  Str out = frame_to_str(in.start, in.pad);
  if (in.start != in.end) {
    out += '-';
    out += frame_to_str(in.end, in.pad);
  }

  //DJV_DBG_PRN("out = " << out);
  return out;
}

Seq str_to_seq(const Str & in)
{
  //DJV_DBG_OBJ2(_DBG, "str_to_seq");
  //DJV_DBG_PRN("in = " << in);

  Seq out;
  const Str::size_type i = in.find('-');
  if (i != Str::npos) {
  
    const Str start(in, 0, i), end(in, i + 1, in.size() - (i + 1));
    //DJV_DBG_PRN("start = " << start);
    //DJV_DBG_PRN("end = " << end);
    
    int pad = 0;
    out.start = str_to_frame(start, &pad);
    out.pad = pad;
    out.end = str_to_frame(end, &pad);
    out.pad = math::max(pad, out.pad);
    
  } else {
  
    out.start = out.end = str_to_frame(in, &out.pad);
  }
  
  //DJV_DBG_PRN("out = " << out.start << "-" << out.end << ":" << out.pad);
  return out;
}

Str frame_to_str(int in, int pad)
{
  char format [] = "%00d";
  static const int format_max = 9;
  format[2] = char('0' + math::min(pad, format_max));
  
  char tmp [str::cstr_size];
  const int size = SNPRINTF(tmp, str::cstr_size, format, in);
  return Str(tmp, size);
}

bool serialize(List<Str> * in, File * out)
{
  Str tmp;
  if (! str::serialize(in, &tmp)) return false;
  out->set(tmp);
  return true;
}

List<Str> serialize(const File & in)
{
  return List<Str>() << in;
}

const List<Str> & type_label()
{
  static const List<Str> data = List<Str>() << "File" << "Seq" << "Dir";
  return data;
}

Str size_label(uint64_t in)
{
  if (in > mem::terabyte)
    return str::Format("%1TB").arg(in / double(mem::terabyte), 0, -2);
  else if (in > mem::gigabyte)
    return str::Format("%1GB").arg(in / double(mem::gigabyte), 0, -2);
  else if (in > mem::megabyte)
    return str::Format("%1MB").arg(in / double(mem::megabyte), 0, -2);
  else
    return str::Format("%1KB").arg(in / double(mem::kilobyte), 0, -2);
}

const Str & perm_label(int in)
{
  static const Str data [] = { "", "r", "w", "rw", "x", "rx", "wx", "rwx" };
  return data[in];
}

const List<Str> & sort_label()
{
  static const List<Str> data = List<Str>() << "Name" << "Type" << "Size" <<
    "User" << "Permissions" << "Time";
  return data;
}

bool operator == (const File & a, const File & b) { return a.get() == b.get(); }
bool operator != (const File & a, const File & b) { return a.get() != b.get(); }

Dbg & operator << (Dbg & dbg, const Seq & in)
  { return dbg << str::Format("%1-%2:%3").arg(in.start).arg(in.end).
      arg(in.pad); }

Dbg & operator << (Dbg & dbg, const File & in)
  { return dbg << in.get(); }

}
}

