/*------------------------------------------------------------------------------
 lib/djv_base/vec_fnc_.h

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#include <djv_base/math.h>

namespace djv {
namespace vec {

template<typename T, int D>
inline Vec<T, D> min(const Vec<T, D> & a, const Vec<T, D> & b)
{
  Vec<T, D> out;
  for (int i = 0; i < D; ++i)
    out.e[i] = math::min(a.e[i], b.e[i]);
  return out;
}

template<typename T, int D>
inline Vec<T, D> max(const Vec<T, D> & a, const Vec<T, D> & b)
{
  Vec<T, D> out;
  for (int i = 0; i < D; ++i)
    out.e[i] = math::max(a.e[i], b.e[i]);
  return out;
}

template<typename T, int D>
inline Vec<T, D> clamp(const Vec<T, D> & in, const Vec<T, D> & min,
  const Vec<T, D> & max
) {
  Vec<T, D> out;
  for (int i = 0; i < D; ++i)
    out.e[i] = math::clamp(in.e[i], min.e[i], max.e[i]);
  return out;
}

template <typename T, int D>
inline Vec<T, D> normalize(const Vec<T, D> & in)
{
  Vec<T, D> out;
  const double l = length(in);
  if (l != 0.0)
    for (int i = 0; i < D; ++i)
      out.e[i] = T(in.e[i] / l);
  return out;
}

template <typename T, int D>
inline double dot(const Vec<T, D> & a, const Vec<T, D> & b)
{
  double out = 0.0;
  for (int i = 0; i < D; ++i)
    out += a.e[i] * b.e[i];
  return out;
}

template <typename T, int D>
inline double length(const Vec<T, D> & in)
{
  double out = 0.0;
  for (int i = 0; i < D; ++i)
    out += in.e[i] * in.e[i];
  return math::sqrt(out);
}

template<typename T, int D>
inline bool size_valid(const Vec<T, D> & in)
{
  return in > Vec<T, D>(0);
}

template<typename T>
inline double aspect(const Vec2<T> & in)
{
  return in.e[1] != T(0) ? (in.e[0] / double(in.e[1])) : 0.0;
}

template <typename T>
inline Vec2<T> swap(const Vec2<T> & in)
{
  return Vec2<T>(in.y, in.x);
}

template <typename T, int D>
inline Vec<T, D> operator + (const Vec<T, D> & a, const Vec<T, D> & b)
{
  Vec<T, D> out;
  for (int i = 0; i < D; ++i)
    out.e[i] = a.e[i] + b.e[i];
  return out;
}

template <typename T, int D>
inline Vec<T, D> operator - (const Vec<T, D> & a, const Vec<T, D> & b)
{
  Vec<T, D> out;
  for (int i = 0; i < D; ++i)
    out.e[i] = a.e[i] - b.e[i];
  return out;
}

template <typename T, int D>
inline Vec<T, D> operator * (const Vec<T, D> & a, const Vec<T, D> & b)
{
  Vec<T, D> out;
  for (int i = 0; i < D; ++i)
    out.e[i] = a.e[i] * b.e[i];
  return out;
}

template <typename T, int D>
inline Vec<T, D> operator / (const Vec<T, D> & a, const Vec<T, D> & b)
{
  Vec<T, D> out;
  for (int i = 0; i < D; ++i)
    out.e[i] = a.e[i] / b.e[i];
  return out;
}

template <typename T, int D>
inline Vec<T, D> operator + (const Vec<T, D> & a, T b)
{
  Vec<T, D> out;
  for (int i = 0; i < D; ++i)
    out.e[i] = a.e[i] + b;
  return out;
}

template <typename T, int D>
inline Vec<T, D> operator - (const Vec<T, D> & a, T b)
{
  Vec<T, D> out;
  for (int i = 0; i < D; ++i)
    out.e[i] = a.e[i] - b;
  return out;
}

template <typename T, int D>
inline Vec<T, D> operator * (const Vec<T, D> & a, T b)
{
  Vec<T, D> out;
  for (int i = 0; i < D; ++i)
    out.e[i] = a.e[i] * b;
  return out;
}

template <typename T, int D>
inline Vec<T, D> operator / (const Vec<T, D> & a, T b)
{
  Vec<T, D> out;
  for (int i = 0; i < D; ++i)
    out.e[i] = a.e[i] / b;
  return out;
}

template <typename T, int D>
inline Vec<T, D> operator - (const Vec<T, D> & in)
{
  Vec<T, D> out;
  for (int i = 0; i < D; ++i)
    out.e[i] = -in.e[i];
  return out;
}

template <typename T, int D>
inline bool operator == (const Vec<T, D> & a, const Vec<T, D> & b)
{
  for (int i = 0; i < D; ++i)
    if (a.e[i] != b.e[i]) return false;
  return true;
}

template <typename T, int D>
inline bool operator != (const Vec<T, D> & a, const Vec<T, D> & b)
  { return ! (a == b); }

template <typename T, int D>
inline bool operator < (const Vec<T, D> & a, const Vec<T, D> & b)
{
  for (int i = 0; i < D; ++i)
    if (a.e[i] >= b.e[i]) return false;
  return true;
}

template <typename T, int D>
inline bool operator <= (const Vec<T, D> & a, const Vec<T, D> & b)
{
  for (int i = 0; i < D; ++i)
    if (a.e[i] > b.e[i]) return false;
  return true;
}

template <typename T, int D>
inline bool operator > (const Vec<T, D> & a, const Vec<T, D> & b)
{
  for (int i = 0; i < D; ++i)
    if (a.e[i] <= b.e[i]) return false;
  return true;
}

template <typename T, int D>
inline bool operator >= (const Vec<T, D> & a, const Vec<T, D> & b)
{
  for (int i = 0; i < D; ++i)
    if (a.e[i] < b.e[i]) return false;
  return true;
}


template <int D>
inline Vec<int, D> to_int(const Vec<double, D> & in)
{
  Vec<int, D> out;
  for (int i = 0; i < D; ++i)
    out.e[i] = int(in.e[i]);
  return out;
}

template <int D>
inline Vec<double, D> to_float(const Vec<int, D> & in)
{
  Vec<double, D> out;
  for (int i = 0; i < D; ++i)
    out.e[i] = in.e[i];
  return out;
}

template <int D>
inline Vec<int, D> ceil(const Vec<double, D> & in)
{
  Vec<int, D> out;
  for (int i = 0; i < D; ++i)
    out.e[i] = math::ceil(in.e[i]);
  return out;
}

template <int D>
inline Vec<int, D> floor(const Vec<double, D> & in)
{
  Vec<int, D> out;
  for (int i = 0; i < D; ++i)
    out.e[i] = math::floor(in.e[i]);
  return out;
}

template <typename T, int D>
inline bool serialize(List<Str> * in, Vec<T, D> * out)
{
  for (int i = 0; i < D; ++i)
    if (! str::serialize(in, &out->e[i])) return false;
  return true;
}

template <typename T, int D>
inline List<Str> serialize(const Vec<T, D> & in)
{
  List<Str> out;
  for (int i = 0; i < D; ++i)
    out << str::serialize(in.e[i]);
  return out;
}

template <typename T, int D>
inline Dbg & operator << (Dbg & dbg, const Vec<T, D> & in)
  { return dbg << serialize(in); }

}
}

