/*------------------------------------------------------------------------------
 lib/djv_image/gl_data.h
 
 OpenGL Image Data

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#ifndef DJV_IMAGE_GL_DATA_H
#define DJV_IMAGE_GL_DATA_H

#include <djv_base/box.h>
#include <djv_image/data.h>
#include <djv_image/color.h>
#include <djv_image/profile.h>
#include <djv_image/gl.h>

namespace djv {
namespace image {
namespace gl {
namespace data {

// Drawing options.

struct Display {

  image::Data lut;

  struct Color {
    Color() : brightness(1), contrast(1), saturation(1) {}
    double brightness, contrast, saturation;
  } color;
  
  struct Levels {
    Levels() : gamma(1) { in[0] = out[0] = 0; in[1] = out[1] = 1; }
    Levels(const Levels & _in) { *this = _in; }
    double in [2], gamma, out [2];
    Levels & operator = (const Levels &);
  } levels;
  
  double soft_clip;
  
  Display() : soft_clip(0) {}
};

enum CHANNEL { CHANNEL_DEFAULT, CHANNEL_RED, CHANNEL_GREEN, CHANNEL_BLUE,
  CHANNEL_ALPHA };

enum FILTER { FILTER_LINEAR, FILTER_NEAREST };

struct Options {

  Options() :
    mirror_x(false), mirror_y(false), rotate(0.0), channel(CHANNEL(0)),
    filter(FILTER(0))
  {}

  bool mirror_x, mirror_y;
  Vec2i position;
  double rotate;
  Vec2i scale;
  profile::Profile profile;
  Display display;
  CHANNEL channel;
  FILTER filter;
  image::Color background;
};

// Image data.

class Data {
public:

  Data(const image::Data * = 0);
  ~Data();
  
  void set(const image::Data *);
  const image::Data * get() const { return _data; }

  void draw(const Options & = Options()) const throw (Error);
  
  static void read(image::Data *);
  static void read(image::Data *, const Box2i &);

private:
  
  const image::Data * _data;
  class _Texture;
  class _Fixed;
  class _Shader;
  std::auto_ptr<_Texture> _texture;
  std::auto_ptr<_Fixed> _fixed;
  std::auto_ptr<_Shader> _shader;
  Options _options;
};

// Functions.

void copy(const image::Data &, image::Data *,
  const Options & = Options()) throw (Error);
void copy(const Data &, void * context_share, image::Data *,
  const Options & = Options()) throw (Error);
void copy(const Data &, const pbuffer::Pbuffer &, image::Data *,
  const Options & = Options()) throw (Error);

enum HISTOGRAM { HISTOGRAM_256, HISTOGRAM_1024 };
void histogram(const Data &, void * context_share, HISTOGRAM, image::Data *,
  Color * min, Color * max, const Options & = Options()) throw (Error);
int histogram(HISTOGRAM);

void gradient(image::Data *) throw (Error);

Mtx4<GLfloat> color_mtx(const Display::Color &);
image::Data color_levels(const Display::Levels &, double soft_clip);

// Comparison.

// XXX LUT data is not compared.

bool operator == (const Display &, const Display &);
bool operator == (const Display::Color &, const Display::Color &);
bool operator == (const Display::Levels &, const Display::Levels &);
bool operator != (const Display &, const Display &);
bool operator != (const Display::Color &, const Display::Color &);
bool operator != (const Display::Levels &, const Display::Levels &);

// Conversion.

// XXX LUT is not serialized.

bool serialize(List<Str> *, Display *);
bool serialize(List<Str> *, Display::Color *);
bool serialize(List<Str> *, Display::Levels *);
List<Str> serialize(const Display &);
List<Str> serialize(const Display::Color &);
List<Str> serialize(const Display::Levels &);

// Labels.

const List<Str> & channel_label();
const List<Str> & filter_label();
const List<Str> & histogram_label();

// Debugging.

Dbg & operator << (Dbg &, const Display &);
Dbg & operator << (Dbg &, CHANNEL);
Dbg & operator << (Dbg &, FILTER);

}
}
}
}

#endif
