/*------------------------------------------------------------------------------
 lib/djv_image/gl_data_texture.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_image/gl_data_texture.cxx"

#include "_gl_data.h"

namespace djv {
namespace image {
namespace gl {
namespace data {

namespace {

const Str
  error_namespace = "OpenGL",
  error_texture_max = "Maximum Texture Size: %1";

GLint gl_filter(FILTER in)
{
  switch (in) {
    case FILTER_NEAREST: return GL_NEAREST;
    case FILTER_LINEAR: return GL_LINEAR;
  }
  return 0;
}

}

Data::_Texture::_Texture() :
  _filter(FILTER(0)),
  _data(0), _init(false),
  _texture(0), _texture_init(false)
{
  DJV_DBG_OBJ2(_DBG, "Data::_Texture::_Texture");
}

Data::_Texture::~_Texture()
{
  DJV_DBG_OBJ2(_DBG, "Data::_Texture::~_Texture");
  
  if (_texture) {
    DJV_DBG_PRN("delete texture");
    DJV_DBG_GL(glDeleteTextures(1, &_texture));
  }
}

void Data::_Texture::set(const image::Data * in)
{
  DJV_DBG_OBJ2(_DBG, "Data::_Texture::set");
  if (in) DJV_DBG_PRN("in = " << *in);
  DJV_DBG_PRN("prev = " << _info);
  
  _data = in;
  if (_data) {
    if (_data->info() != _info) {
      _info = _data->info();
      _init = false;
    }
  } else {
    _info = image::data::Info();
    _init = false;
  }
  _texture_init = false;
}

void Data::_Texture::draw(const Options & in) throw (Error)
{
  if (! _data) return;
  
  DJV_DBG_OBJ2(_DBG, "Data::_Texture::draw");
  DJV_DBG_PRN("scale = " << in.scale);
  
  if (in.filter != _filter) {
    _filter = in.filter;
    _init = false;
    _texture_init = false;
  }
  
  const image::data::Info & info = _data->info();
  DJV_DBG_PRN("data = " << info);
  
  bool mirror_x = in.mirror_x, mirror_y = in.mirror_y;
  if (info.mirror_x) mirror_x = ! mirror_x;
  if (info.mirror_y) mirror_y = ! mirror_y;
  DJV_DBG_PRN("mirror = " << mirror_x << " " << mirror_y);

  // Initialize texture.
  
  GLenum target = GL_TEXTURE_2D;
  if (GLEW_ARB_fragment_shader) {
    if (GLEW_ARB_texture_rectangle) target = GL_TEXTURE_RECTANGLE_ARB;
    DJV_DBG_GL(glActiveTexture(GL_TEXTURE0));
  }

  if (! _init) {
    DJV_DBG_PRN("init");

    GLint max = 0;
    DJV_DBG_GL(glGetIntegerv(GL_MAX_TEXTURE_SIZE, &max));
    DJV_DBG_PRN("texture max = " << max);

    _texture_size = _data->size();
    if (GL_TEXTURE_2D == target)
      _texture_size = Vec2i(
        math::to_pow2(_texture_size.x),
        math::to_pow2(_texture_size.y));
    if (_texture_size.x > max || _texture_size.y > max)
      throw Error(error_namespace, str::Format(error_texture_max).
        arg(int(max)));
    DJV_DBG_PRN("texture size = " << _texture_size);

    if (_texture) {
      DJV_DBG_PRN("delete texture");
      DJV_DBG_GL(glDeleteTextures(1, &_texture));
    }
    DJV_DBG_GL(glGenTextures(1, &_texture));
    DJV_DBG_GL(glBindTexture(target, _texture));
    DJV_DBG_GL(glTexParameteri(target, GL_TEXTURE_WRAP_S, GL_CLAMP));
    DJV_DBG_GL(glTexParameteri(target, GL_TEXTURE_WRAP_T, GL_CLAMP));
    DJV_DBG_GL(glTexParameteri(target,
      GL_TEXTURE_MAG_FILTER, gl_filter(_filter)));
    DJV_DBG_GL(glTexParameteri(target,
      GL_TEXTURE_MIN_FILTER, gl_filter(_filter)));
    DJV_DBG_GL(glTexImage2D(
      target, 0,
      GLEW_ARB_fragment_shader && // XXX
        (pixel::F16 == pixel::type(info.pixel) ||
         pixel::F32 == pixel::type(info.pixel)) ?
        GL_FLOAT_RGBA_NV : GL_RGBA,
      _texture_size.x, _texture_size.y, 0,
      gl::format(info.pixel, info.bgr), gl::type(info.pixel),
      0));

    _init = true;
  }

  if (_texture) { DJV_DBG_GL(glBindTexture(target, _texture)); }
  
  // Download texture.
  
  if (! _texture_init) {
    DJV_DBG_PRN("texture init");

    DJV_DBG_GL(glPixelStorei(GL_UNPACK_ALIGNMENT, info.align));
    DJV_DBG_GL(glPixelStorei(GL_UNPACK_SWAP_BYTES,
      info.endian != mem::endian()));
    DJV_DBG_GL(glTexSubImage2D(
      target, 0,
      0, 0, _data->w(), _data->h(),
      gl::format(info.pixel, info.bgr), gl::type(info.pixel),
      _data->data()));

    _texture_init = true;
  }
  
  // Draw.
  
  const int proxy_scale = image::data::proxy_scale(_data ? info.proxy : 0);
  const int w = _data->w() * proxy_scale, h = _data->h() * proxy_scale;

  double u [] = { 0, 0 }, v [] = { 0, 0 };
  u[! mirror_x] = _data->w();
  v[! mirror_y] = _data->h();
  if (GL_TEXTURE_2D == target) {
    u[! mirror_x] /= double(_texture_size.x);
    v[! mirror_y] /= double(_texture_size.y);
  }
  DJV_DBG_PRN("u = " << u[0] << " " << u[1]);
  DJV_DBG_PRN("v = " << v[0] << " " << v[1]);
  
  DJV_DBG_GL(glEnable(target));
  DJV_DBG_GL(glPushMatrix());
  DJV_DBG_GL(glTranslated(in.position.x, in.position.y, 0));
  DJV_DBG_GL(glRotated(-in.rotate, 0, 0, 1));
  if (size_valid(in.scale)) {
    DJV_DBG_GL(glScaled(in.scale.x / double(w), in.scale.y / double(h), 1.0));
  }
  DJV_DBG_GL(glBegin(GL_QUADS));
  DJV_DBG_GL(glTexCoord2d(u[0], v[0]));
  DJV_DBG_GL(glVertex2i(0, 0));
  DJV_DBG_GL(glTexCoord2d(u[0], v[1]));
  DJV_DBG_GL(glVertex2i(0, h));
  DJV_DBG_GL(glTexCoord2d(u[1], v[1]));
  DJV_DBG_GL(glVertex2i(w, h));
  DJV_DBG_GL(glTexCoord2d(u[1], v[0]));
  DJV_DBG_GL(glVertex2i(w, 0));
  DJV_DBG_GL(glEnd());
  DJV_DBG_GL(glPopMatrix());
}

}
}
}
}

