/*------------------------------------------------------------------------------
 lib/djv_image/io_cineon_header.cxx
 
 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_image/io_cineon_header.cxx"

#include "io_cineon.h"

namespace djv {
namespace image {
namespace io {
namespace cineon {

namespace {

enum ORIENT {
  ORIENT_LEFT_RIGHT_TOP_BOTTOM,
  ORIENT_LEFT_RIGHT_BOTTOM_TOP,
  ORIENT_RIGHT_LEFT_TOP_BOTTOM,
  ORIENT_RIGHT_LEFT_BOTTOM_TOP,
  ORIENT_TOP_BOTTOM_LEFT_RIGHT,
  ORIENT_TOP_BOTTOM_RIGHT_LEFT,
  ORIENT_BOTTOM_TOP_LEFT_RIGHT,
  ORIENT_BOTTOM_TOP_RIGHT_LEFT
};

enum DESCRIPTOR {
  DESCRIPTOR_L = 0,
  DESCRIPTOR_R_PRINT = 1,
  DESCRIPTOR_G_PRINT = 2,
  DESCRIPTOR_B_PRINT = 3,
  DESCRIPTOR_R_CCIR_XA11 = 4,
  DESCRIPTOR_G_CCIR_XA11 = 5,
  DESCRIPTOR_B_CCIR_XA11 = 6
};

struct Header {

  struct File {
    uint32_t magic;
    uint32_t image_offset;
    uint32_t header_size;
    uint32_t industry_header_size;
    uint32_t user_header_size;
    uint32_t size;
    char version [8];
    char name [100];
    char time [24];
    uint8_t _pad [36];
  } file;
  
  struct Image {
    uint8_t orient;
    uint8_t channels;
    uint8_t _pad [2];

    struct Channel {
      uint8_t descriptor [2];
      uint8_t bit_depth;
      uint8_t _pad;
      uint32_t size [2];
      float low_data;
      float low_quantity;
      float high_data;
      float high_quantity;
    } channel [8];
    
    float white [2];
    float red [2];
    float green [2];
    float blue [2];
    char label [200];
    uint8_t _pad2 [28];
    uint8_t interleave;
    uint8_t packing;
    uint8_t data_sign;
    uint8_t data_sense;
    uint32_t line_padding;
    uint32_t channel_padding;
    uint8_t _pad3 [20];
  } image;
  
  struct Source {
    int32_t offset [2];
    char file [100];
    char time [24];
    char input_device [64];
    char input_model [32];
    char input_serial [32];
    float input_pitch [2];
    float gamma;
    char _pad [40];
  } source;
  
  struct Film {
    uint8_t id;
    uint8_t type;
    uint8_t offset;
    uint8_t _pad;
    uint32_t prefix;
    uint32_t count;
    char format [32];
    uint32_t frame;
    float frame_rate;
    char frame_id [32];
    char slate [200];
    char _pad2 [740];
  } film;
};

const List<Str> & tag_label()
{
  static const List<Str> data = List<Str>() <<
    "Source Offset" <<
    "Source File" <<
    "Source Time" <<
    "Source Input Device" <<
    "Source Input Model" <<
    "Source Input Serial" <<
    "Source Input Pitch" <<
    "Source Gamma" <<
    "Film Format" <<
    "Film Frame" <<
    "Film Frame Rate" <<
    "Film Frame ID" <<
    "Film Slate";
  return data;
}

inline void zero(int32_t * in) { *((uint32_t *)in) = 0x80000000; }
inline void zero(float * in) { *((uint32_t *)in) = 0x7F800000; }
inline void zero(char * in, int size) { mem::zero(in, size); }

inline bool valid(const uint8_t * in) { return *in != 0xff; }
inline bool valid(const uint16_t * in) { return *in != 0xffff; }

// XXX Catch uninitialized values.

const int32_t _int_max = 1000000;
const float _float_max = 1000000.0;
const char _char_min = 32;
const char _char_max = 126;

inline bool valid(const uint32_t * in)
{
  return
    *in != 0xffffffff &&
    *in < uint32_t(_int_max);
}

inline bool valid(const int32_t * in)
{
  return
    *in != int32_t(0x80000000) &&
    *in > -_int_max && *in < _int_max;
}

inline bool valid(const float * in)
{
  return
    *((uint32_t *)in) != 0x7F800000 &&
    *in > -_float_max && *in < _float_max;
}

inline bool valid(const char * in, int size)
{
  const char * p = in;
  const char * const end = p + size;
  for (; *p && p < end; ++p)
    if (*p < _char_min || *p > _char_max)
      return false;
  return size ? (in[0] != 0) : false;
}

inline Str _str(const char * in, int size)
{
  const char * p = in;
  const char * const end = p + size;
  for (; *p && p < end; ++p) ;
  return Str(in, p - in);
}

Str dbg(uint8_t in)
  { return str::int_to_str(in); }
  //{ return valid(&in) ? str::int_to_str(in) : "-"; }
Str dbg(uint16_t in)
  { return str::int_to_str(in); }
  //{ return valid(&in) ? str::int_to_str(in) : "-"; }
Str dbg(uint32_t in)
  { return str::int_to_str(in); }
  //{ return valid(&in) ? str::int_to_str(in) : "-"; }
  //{ return dbg_u32(in); }
Str dbg(int32_t in)
  { return str::int_to_str(in); }
  //{ return valid(&in) ? str::int_to_str(in) : "-"; }
  //{ return dbg_u32(in); }
Str dbg(float in)
  { return str::float_to_str(in); }
  //{ return valid(&in) ? str::float_to_str(in) : "-"; }
  //{ return dbg_u32(in); }
Str dbg(const char * in, int size)
  { return Str(in, size); }
  //{ return valid(in, size) ? Str(in, size) : "-"; }

void dbg(const Header & in)
{
  DJV_DBG_OBJ2(_DBG, "dbg");
  
  DJV_DBG_PRN("file");
  DJV_DBG_PRN("  image offset = " << dbg(in.file.image_offset));
  DJV_DBG_PRN("  header size = " << dbg(in.file.header_size));
  DJV_DBG_PRN("  industry header size = " << dbg(in.file.industry_header_size));
  DJV_DBG_PRN("  user header size = " << dbg(in.file.user_header_size));
  DJV_DBG_PRN("  size = " << dbg(in.file.size));
  DJV_DBG_PRN("  version = " << dbg(in.file.version, 8));
  DJV_DBG_PRN("  name = " << dbg(in.file.name, 100));
  DJV_DBG_PRN("  time = " << dbg(in.file.time, 24));

  DJV_DBG_PRN("image");
  DJV_DBG_PRN("  orient = " << dbg(in.image.orient));
  DJV_DBG_PRN("  channels = " << dbg(in.image.channels));
  for (int i = 0; i < in.image.channels; ++i) {
    DJV_DBG_PRN("  channel " << i);
    DJV_DBG_PRN("    descriptor = " <<
      dbg(in.image.channel[i].descriptor[0]) << " " <<
      dbg(in.image.channel[i].descriptor[1]));
    DJV_DBG_PRN("    bit depth = " << dbg(in.image.channel[i].bit_depth));
    DJV_DBG_PRN("    size = " << dbg(in.image.channel[i].size[0]) << " " <<
      dbg(in.image.channel[i].size[1]));
    DJV_DBG_PRN("    low data = " << dbg(in.image.channel[i].low_data));
    DJV_DBG_PRN("    low quantity = " << dbg(in.image.channel[i].low_quantity));
    DJV_DBG_PRN("    high data = " << dbg(in.image.channel[i].high_data));
    DJV_DBG_PRN("    high quantity = " <<
      dbg(in.image.channel[i].high_quantity));
  }
  DJV_DBG_PRN("  white = " << dbg(in.image.white[0]) << " " <<
    dbg(in.image.white[1]));
  DJV_DBG_PRN("  red = " << dbg(in.image.red[0]) << " " <<
    dbg(in.image.red[1]));
  DJV_DBG_PRN("  green = " << dbg(in.image.green[0]) << " " <<
    dbg(in.image.green[1]));
  DJV_DBG_PRN("  blue = " << dbg(in.image.blue[0]) << " " <<
    dbg(in.image.blue[1]));
  DJV_DBG_PRN("  label = " << dbg(in.image.label, 200));
  DJV_DBG_PRN("  interleave = " << dbg(in.image.interleave));
  DJV_DBG_PRN("  packing = " << dbg(in.image.packing));
  DJV_DBG_PRN("  data sign = " << dbg(in.image.data_sign));
  DJV_DBG_PRN("  data sense = " << dbg(in.image.data_sense));
  DJV_DBG_PRN("  line padding = " << dbg(in.image.line_padding));
  DJV_DBG_PRN("  channel padding = " << dbg(in.image.channel_padding));
  
  DJV_DBG_PRN("source");
  DJV_DBG_PRN("  offset = " << dbg(in.source.offset[0]) << " " <<
    dbg(in.source.offset[1]));
  DJV_DBG_PRN("  file = " << dbg(in.source.file, 100));
  DJV_DBG_PRN("  time = " << dbg(in.source.time, 24));
  DJV_DBG_PRN("  input device = " << dbg(in.source.input_device, 64));
  DJV_DBG_PRN("  input model = " << dbg(in.source.input_model, 32));
  DJV_DBG_PRN("  input serial = " << dbg(in.source.input_serial, 32));
  DJV_DBG_PRN("  input pitch = " << dbg(in.source.input_pitch[0]) << " " <<
    dbg(in.source.input_pitch[1]));
  DJV_DBG_PRN("  gamma = " << dbg(in.source.gamma));

  DJV_DBG_PRN("film");
  DJV_DBG_PRN("  id = " << dbg(in.film.id));
  DJV_DBG_PRN("  type = " << dbg(in.film.type));
  DJV_DBG_PRN("  offset = " << dbg(in.film.offset));
  DJV_DBG_PRN("  prefix = " << dbg(in.film.prefix));
  DJV_DBG_PRN("  count = " << dbg(in.film.count));
  DJV_DBG_PRN("  format = " << dbg(in.film.format, 32));
  DJV_DBG_PRN("  frame = " << dbg(in.film.frame));
  DJV_DBG_PRN("  frame rate = " << dbg(in.film.frame_rate));
  DJV_DBG_PRN("  frame id = " << dbg(in.film.frame_id, 32));
  DJV_DBG_PRN("  slate = " << dbg(in.film.slate, 200));
}

void _file(const Str & file, const image::Info & info, Header::File * out)
{
  const List<Str> & tag_label = tag::label();
  
  out->image_offset = 2048;
  out->header_size = 1024;
  out->industry_header_size = 1024;
  out->user_header_size = 0;
  
  zero(out->version, 8);
   
  zero(out->name, 100);
  str::ascii(file, out->name, 100, false);

  zero(out->time, 24);
  str::ascii(info.tag[tag_label[tag::TIME]], out->time, 24, false);
}

void _file(const Header::File & in, image::Info * info)
{
  const List<Str> & tag_label = tag::label();
  
  if (valid(in.time, 24))
    info->tag[tag_label[tag::TIME]] = _str(in.time, 24);
}

void _image(const image::Info & info, dpx::COLOR color, Header::Image * out)
{
  for (uint i = 0; i < 8; ++i) {
    zero(&out->channel[i].low_data);
    zero(&out->channel[i].low_quantity);
    zero(&out->channel[i].high_data);
    zero(&out->channel[i].high_quantity);
  }

  out->orient = ORIENT_LEFT_RIGHT_TOP_BOTTOM;

  out->channels = 3;  
  if (dpx::COLOR_PRINT == color) {
    out->channel[0].descriptor[1] = DESCRIPTOR_R_PRINT;
    out->channel[1].descriptor[1] = DESCRIPTOR_G_PRINT;
    out->channel[2].descriptor[1] = DESCRIPTOR_B_PRINT;
  } else {
    out->channel[0].descriptor[1] = DESCRIPTOR_L;
    out->channel[1].descriptor[1] = DESCRIPTOR_L;
    out->channel[2].descriptor[1] = DESCRIPTOR_L;
  }
  
  const int bit_depth = 10;
  for (int i = 0; i < out->channels; ++i) {
  
    out->channel[i].descriptor[0] = 0;
    out->channel[i].bit_depth = bit_depth;
    out->channel[i].size[0] = info.size.x;
    out->channel[i].size[1] = info.size.y;

    out->channel[i].low_data = 0;
    switch (bit_depth) {
      case 8: out->channel[i].high_data = 255; break;
      case 10: out->channel[i].high_data = 1023; break;
      case 12: out->channel[i].high_data = 4095; break;
      case 16: out->channel[i].high_data = 65535; break;
    }
  }
  
  out->interleave = 0;
  out->packing = 5;
  out->data_sign = 0;
  out->data_sense = 0;
  out->line_padding = 0;
  out->channel_padding = 0;
}

void _image(const Header::Image & in, image::Info * info, bool * print)
  throw (Error)
{
  switch (in.orient) {
    case ORIENT_LEFT_RIGHT_TOP_BOTTOM: info->mirror_y = true; break;
    case ORIENT_LEFT_RIGHT_BOTTOM_TOP: break;
    case ORIENT_RIGHT_LEFT_TOP_BOTTOM: info->mirror_x = true; break;
    case ORIENT_RIGHT_LEFT_BOTTOM_TOP:
      info->mirror_x = info->mirror_y = true; break;
    case ORIENT_TOP_BOTTOM_LEFT_RIGHT:
    case ORIENT_TOP_BOTTOM_RIGHT_LEFT:
    case ORIENT_BOTTOM_TOP_LEFT_RIGHT:
    case ORIENT_BOTTOM_TOP_RIGHT_LEFT: break;
  }
  
  if (! in.channels)
    throw Error(plugin_name, error_unsupported);
  int i = 1;
  for (; i < in.channels; ++i) {
    if (
      (in.channel[i].size[0] != in.channel[0].size[0]) ||
      (in.channel[i].size[1] != in.channel[0].size[1])
    )
      break;
    if (in.channel[i].bit_depth != in.channel[0].bit_depth)
      break;
  }
  if (i < in.channels)
    throw Error(plugin_name, error_unsupported);
  
  int pixel = -1;
  switch (in.channels) {
    case 3:
      switch (in.channel[0].bit_depth) {
        case 10: pixel = pixel::RGB_U10; break;
      }
      break;
  }
  if (-1 == pixel)
    throw Error(plugin_name, error_unsupported);
  info->pixel = pixel::PIXEL(pixel);
  
  info->size = Vec2i(in.channel[0].size[0], in.channel[0].size[1]);
      
  if (valid(&in.line_padding) && in.line_padding)
    throw Error(plugin_name, error_unsupported);
  if (valid(&in.channel_padding) && in.channel_padding)
    throw Error(plugin_name, error_unsupported);
  
  if (print)
    *print = DESCRIPTOR_R_PRINT == in.channel[0].descriptor[1];
}

void _source(const image::Info & info, Header::Source * out)
{
  const List<Str> & cineon_tag_label = cineon::tag_label();
  Str tag;
  
  zero(&out->offset[0]);
  zero(&out->offset[1]);
  tag = info.tag[cineon_tag_label[TAG_SOURCE_OFFSET]];
  if (tag.size()) {
    const List<Str> tmp = str::split(tag, " ");
    if (2 == tmp.size()) {
      out->offset[0] = str::str_to_int(tmp[0]);
      out->offset[1] = str::str_to_int(tmp[1]);
    }
  }
  
  zero(out->file, 100);
  tag = info.tag[cineon_tag_label[TAG_SOURCE_FILE]];
  if (tag.size()) str::ascii(tag, out->file, 100, false);

  zero(out->time, 24);
  tag = info.tag[cineon_tag_label[TAG_SOURCE_TIME]];
  if (tag.size()) str::ascii(tag, out->time, 24, false);

  zero(out->input_device, 64);
  tag = info.tag[cineon_tag_label[TAG_SOURCE_INPUT_DEVICE]];
  if (tag.size()) str::ascii(tag, out->input_device, 64, false);

  zero(out->input_model, 32);
  tag = info.tag[cineon_tag_label[TAG_SOURCE_INPUT_MODEL]];
  if (tag.size()) str::ascii(tag, out->input_model, 32, false);

  zero(out->input_serial, 32);
  tag = info.tag[cineon_tag_label[TAG_SOURCE_INPUT_SERIAL]];
  if (tag.size()) str::ascii(tag, out->input_serial, 32, false);
  
  zero(&out->input_pitch[0]);
  zero(&out->input_pitch[1]);
  tag = info.tag[cineon_tag_label[TAG_SOURCE_INPUT_PITCH]];
  if (tag.size()) {
    const List<Str> tmp = str::split(tag, " ");
    if (2 == tmp.size()) {
      out->offset[0] = str::str_to_int(tmp[0]);
      out->offset[1] = str::str_to_int(tmp[1]);
    }
  }
  
  zero(&out->gamma);
  tag = info.tag[cineon_tag_label[TAG_SOURCE_GAMMA]];
  if (tag.size()) out->gamma = str::str_to_float(tag);
}

void _source(const Header::Source & in, image::Info * info)
{
  const List<Str> & cineon_tag_label = cineon::tag_label();
  
  if (valid(&in.offset[0]) && valid(&in.offset[1]))
    info->tag[cineon_tag_label[TAG_SOURCE_OFFSET]] = str::join(List<Str>() <<
      str::int_to_str(in.offset[0]) << str::int_to_str(in.offset[1]), " ");

  if (valid(in.file, 100))
    info->tag[cineon_tag_label[TAG_SOURCE_FILE]] = _str(in.file, 100);
  
  if (valid(in.time, 24))
    info->tag[cineon_tag_label[TAG_SOURCE_TIME]] = _str(in.time, 24);
  
  if (valid(in.input_device, 64))
    info->tag[cineon_tag_label[TAG_SOURCE_INPUT_DEVICE]] =
      _str(in.input_device, 64);
  
  if (valid(in.input_model, 32))
    info->tag[cineon_tag_label[TAG_SOURCE_INPUT_MODEL]] =
      _str(in.input_model, 32);
  
  if (valid(in.input_serial, 32))
    info->tag[cineon_tag_label[TAG_SOURCE_INPUT_SERIAL]] =
      _str(in.input_serial, 32);
  
  if (valid(&in.input_pitch[0]) && valid(&in.input_pitch[1]))
    info->tag[cineon_tag_label[TAG_SOURCE_INPUT_PITCH]] = str::join(
      List<Str>() << str::float_to_str(in.input_pitch[0]) <<
      str::float_to_str(in.input_pitch[1]), " ");
   
  if (valid(&in.gamma))
    info->tag[cineon_tag_label[TAG_SOURCE_GAMMA]] =
      str::float_to_str(in.gamma);
}

void _film(const image::Info & info, Header::Film * out)
{
  const List<Str> & tag_label = tag::label();
  const List<Str> & cineon_tag_label = cineon::tag_label();
  Str tag;

  tag = info.tag[tag_label[tag::KEYCODE]];
  if (tag.size()) {
    int id = 0, type = 0, prefix = 0, count = 0, offset = 0;
    time::str_to_keycode(tag, &id, &type, &prefix, &count, &offset);
    out->id = id;
    out->type = type;
    out->offset = offset;
    out->prefix = prefix;
    out->count = count;
  }
  
  zero(out->format, 32);
  tag = info.tag[cineon_tag_label[TAG_FILM_FORMAT]];
  if (tag.size()) str::ascii(tag, out->format, 32, false);

  tag = info.tag[cineon_tag_label[TAG_FILM_FRAME]];
  if (tag.size()) out->frame = str::str_to_int(tag);

  zero(&out->frame_rate);
  tag = info.tag[cineon_tag_label[TAG_FILM_FRAME_RATE]];
  if (tag.size()) out->frame_rate = str::str_to_float(tag);

  zero(out->frame_id, 32);
  tag = info.tag[cineon_tag_label[TAG_FILM_FRAME_ID]];
  if (tag.size()) str::ascii(tag, out->frame_id, 32, false);

  zero(out->slate, 200);
  tag = info.tag[cineon_tag_label[TAG_FILM_SLATE]];
  if (tag.size()) str::ascii(tag, out->slate, 200, false);
}

void _film(const Header::Film & in, Info * info)
{
  const List<Str> & tag_label = tag::label();
  const List<Str> & cineon_tag_label = cineon::tag_label();
  
  if (valid(&in.id) && valid(&in.type) && valid(&in.offset) &&
    valid(&in.prefix) && valid(&in.count)
  )
    info->tag[tag_label[tag::KEYCODE]] = time::keycode_to_str(
      in.id, in.type, in.prefix, in.count, in.offset);

  if (valid(in.format, 32))
    info->tag[cineon_tag_label[TAG_FILM_FORMAT]] = _str(in.format, 32);
  
  if (valid(&in.frame))
    info->tag[cineon_tag_label[TAG_FILM_FRAME]] =
      str::int_to_str(in.frame);
  
  if (valid(&in.frame_rate) && in.frame_rate >= 0.000001) {
    info->time.speed = time::float_to_speed(in.frame_rate);
    info->tag[cineon_tag_label[TAG_FILM_FRAME_RATE]] =
      str::float_to_str(in.frame_rate);
  }
  
  if (valid(in.frame_id, 32))
    info->tag[cineon_tag_label[TAG_FILM_FRAME_ID]] = _str(in.frame_id, 32);
  
  if (valid(in.slate, 200))
    info->tag[cineon_tag_label[TAG_FILM_SLATE]] = _str(in.slate, 200);
}

void _endian(Header * in)
{
  mem::endian(&in->file.image_offset, 1, 4);
  mem::endian(&in->file.header_size, 1, 4);
  mem::endian(&in->file.industry_header_size, 1, 4);
  mem::endian(&in->file.user_header_size, 1, 4);
  mem::endian(&in->file.size, 1, 4);

  for (uint i = 0; i < 8; ++i) {
    mem::endian(in->image.channel[i].size, 2, 4);
    mem::endian(&in->image.channel[i].low_data, 1, 4);
    mem::endian(&in->image.channel[i].low_quantity, 1, 4);
    mem::endian(&in->image.channel[i].high_data, 1, 4);
    mem::endian(&in->image.channel[i].high_quantity, 1, 4);
  }
  mem::endian(in->image.white, 2, 4);
  mem::endian(in->image.red, 2, 4);
  mem::endian(in->image.green, 2, 4);
  mem::endian(in->image.blue, 2, 4);
  mem::endian(&in->image.line_padding, 1, 4);
  mem::endian(&in->image.channel_padding, 1, 4);
  
  mem::endian(in->source.offset, 2, 4);
  mem::endian(in->source.input_pitch, 2, 4);
  mem::endian(&in->source.gamma, 1, 4);
  
  mem::endian(&in->film.prefix, 1, 4);
  mem::endian(&in->film.count, 1, 4);
  mem::endian(&in->film.frame, 1, 4);
  mem::endian(&in->film.frame_rate, 1, 4);
}

}

void header_set(djv::Io * io, const Str & file, const image::Info & info,
  dpx::COLOR color
) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "header_set");
  DJV_DBG_PRN("info = " << info);
  DJV_DBG_PRN("color = " << color);
  
  Header header;
  mem::set(0xff, &header, sizeof(Header));
  _file(file, info, &header.file);
  _image(info, color, &header.image);
  _source(info, &header.source);
  _film(info, &header.film);
  dbg(header);
  
  const bool endian = mem::endian() != mem::MSB;
  io->endian(endian);
  if (endian) {
    DJV_DBG_PRN("endian");
    _endian(&header);
    header.file.magic = 0xd75f2a80;
  } else {
    header.file.magic = 0x802a5fd7;
  }
  io->set(&header, sizeof(Header));
}

void header_update(djv::Io * io) throw (Error)
{
  const size_t size = io->position();
  io->position(20);
  io->set_u32(size);
}

void header_get(djv::Io * io, Info * info, bool * print) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "header_get");
  
  Header header;
  io->get(&header, sizeof(Header));
  if (0x802a5fd7 == header.file.magic) {
    io->endian(false);
  } else if (0xd75f2a80 == header.file.magic) {
    DJV_DBG_PRN("endian");
    io->endian(true);
    _endian(&header);
    info->endian = mem::_endian;
  } else
    throw Error(plugin_name, error_unrecognized);
  dbg(header);
  
  _file(header.file, info);
  _image(header.image, info, print);
  _source(header.source, info);
  _film(header.film, info);
  
  if (header.file.image_offset)
    io->position(header.file.image_offset);
}

}
}
}
}

