/*------------------------------------------------------------------------------
 lib/djv_image/io_dpx_header.cxx
 
 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_image/io_dpx_header.cxx"

#include "io_dpx.h"

namespace djv {
namespace image {
namespace io {
namespace dpx {

namespace {

enum ORIENT {
  ORIENT_LEFT_RIGHT_TOP_BOTTOM,
  ORIENT_RIGHT_LEFT_TOP_BOTTOM,
  ORIENT_LEFT_RIGHT_BOTTOM_TOP,
  ORIENT_RIGHT_LEFT_BOTTOM_TOP,
  ORIENT_TOP_BOTTOM_LEFT_RIGHT,
  ORIENT_TOP_BOTTOM_RIGHT_LEFT,
  ORIENT_BOTTOM_TOP_LEFT_RIGHT,
  ORIENT_BOTTOM_TOP_RIGHT_LEFT
};

enum DESCRIPTOR {
  DESCRIPTOR_USER = 0,
  DESCRIPTOR_R = 1,
  DESCRIPTOR_G = 2,
  DESCRIPTOR_B = 3,
  DESCRIPTOR_A = 4,
  DESCRIPTOR_L = 6,
  DESCRIPTOR_CHROMA = 7,
  DESCRIPTOR_Z = 8,
  DESCRIPTOR_COMPOSITE_VIDEO = 9,
  DESCRIPTOR_RGB = 50,
  DESCRIPTOR_RGBA = 51,
  DESCRIPTOR_ABGR = 52,
  DESCRIPTOR_CBYCRY = 100,
  DESCRIPTOR_CBYACRYA = 101,
  DESCRIPTOR_CBYCR = 102,
  DESCRIPTOR_CBYCRA = 103,
  DESCRIPTOR_USER_2 = 150,
  DESCRIPTOR_USER_3 = 151,
  DESCRIPTOR_USER_4 = 152,
  DESCRIPTOR_USER_5 = 153,
  DESCRIPTOR_USER_6 = 154,
  DESCRIPTOR_USER_7 = 155,
  DESCRIPTOR_USER_8 = 156
};

enum TRANSFER {
  TRANSFER_USER,
  TRANSFER_PRINT,
  TRANSFER_LINEAR,
  TRANSFER_LOG,
  TRANSFER_VIDEO,
  TRANSFER_SMPTE_274M,
  TRANSFER_ITU_R_709_4,
  TRANSFER_ITU_R_601_5_B_OR_G,
  TRANSFER_ITU_R_601_5_M,
  TRANSFER_NTSC,
  TRANSFER_PAL,
  TRANSFER_Z,
  TRANSFER_Z_HOMOGENEOUS
};

enum COLORIMETRIC {
  COLORIMETRIC_USER = 0,
  COLORIMETRIC_PRINT = 1,
  COLORIMETRIC_VIDEO = 4,
  COLORIMETRIC_SMPTE_274M = 5,
  COLORIMETRIC_ITU_R_709_4 = 6,
  COLORIMETRIC_ITU_R_601_5_B_OR_G = 7,
  COLORIMETRIC_ITU_R_601_5_M = 8,
  COLORIMETRIC_NTSC = 9,
  COLORIMETRIC_PAL = 10
};

enum COMPONENTS { PACK, TYPE_A, TYPE_B };

static const char magic [][5] = { "SDPX", "XPDS" };

struct Header {

  struct File {
    uint32_t magic;
    uint32_t image_offset;
    char version [8];
    uint32_t size;
    uint32_t ditto_key;
    uint32_t header_size;
    uint32_t industry_header_size;
    uint32_t user_header_size;
    char name [100];
    char time [24];
    char creator [100];
    char project [200];
    char copyright [200];
    uint32_t encryption_key;
    uint8_t _pad [104];
  } file;
  
  struct Image {
    uint16_t orient;
    uint16_t elem_size;
    uint32_t size [2];
    
    struct Elem {
      uint32_t data_sign;
      uint32_t low_data;
      float low_quantity;
      uint32_t high_data;
      float high_quantity;
      uint8_t descriptor;
      uint8_t transfer;
      uint8_t colorimetric;
      uint8_t bit_depth;
      uint16_t packing, encoding;
      uint32_t data_offset;
      uint32_t line_padding;
      uint32_t elem_padding;
      char description [32];
    } elem [8];

    uint8_t _pad [52];
  } image;

  struct Source {
    uint32_t offset [2];
    float center [2];
    uint32_t size [2];
    char file [100];
    char time [24];
    char input_device [32];
    char input_serial [32];
    uint16_t border [4];
    uint32_t pixel_aspect [2];
    float scan_size [2];
    uint8_t _pad [20];
  } source;

  struct Film {
    char id [2];
    char type [2];
    char offset [2];
    char prefix [6];
    char count [4];
    char format [32];
    uint32_t frame;
    uint32_t sequence;
    uint32_t hold;
    float frame_rate;
    float shutter;
    char frame_id [32];
    char slate [100];
    uint8_t _pad [56];
  } film;

  struct Tv {
    uint32_t timecode;
    uint32_t user_bits;
    uint8_t interlace;
    uint8_t field, video_signal;
    uint8_t _pad;
    float sample_rate [2];
    float frame_rate;
    float time_offset;
    float gamma;
    float black_level;
    float black_gain;
    float breakpoint;
    float white_level;
    float integration_times;
    uint8_t _pad2 [76];
  } tv;
};

inline void zero(char * in, int size) { mem::zero(in, size); }

bool valid(const uint8_t * in) { return *in != 0xff; }
bool valid(const uint16_t * in) { return *in != 0xffff; }

// XXX Catch uninitialized values.

const int32_t _int_max = 1000000;
const float _float_max = 1000000.0;
const char _char_min = 32;
const char _char_max = 126;

bool valid(const uint32_t * in)
{
  return
    *in != 0xffffffff &&
    *in < uint32_t(_int_max);
}

bool valid(const float * in)
{
  return
    *((uint32_t *)in) != 0xffffffff &&
    *in > -_float_max && *in < _float_max;
}

bool valid(const char * in, int size)
{
  const char * p = in;
  const char * const end = p + size;
  for (; *p && p < end; ++p)
    if (*p < _char_min || *p > _char_max)
      return false;
  return size ? (in[0] != 0) : false;
}

inline Str _str(const char * in, int size)
{
  const char * p = in;
  const char * const end = p + size;
  for (; *p && p < end; ++p) ;
  return Str(in, p - in);
}

void _endian(Header * in)
{
  mem::endian(&in->file.image_offset, 1, 4);
  mem::endian(&in->file.size, 1, 4);
  mem::endian(&in->file.ditto_key, 1, 4);
  mem::endian(&in->file.header_size, 1, 4);
  mem::endian(&in->file.industry_header_size, 1, 4);
  mem::endian(&in->file.user_header_size, 1, 4);
  mem::endian(&in->file.encryption_key, 1, 4);
  
  mem::endian(&in->image.orient, 1, 2);
  mem::endian(&in->image.elem_size, 1, 2);
  mem::endian(in->image.size, 2, 4);
  for (uint i = 0; i < 8; ++i) {
    mem::endian(&in->image.elem[i].data_sign, 1, 4);
    mem::endian(&in->image.elem[i].low_data, 1, 4);
    mem::endian(&in->image.elem[i].low_quantity, 1, 4);
    mem::endian(&in->image.elem[i].high_data, 1, 4);
    mem::endian(&in->image.elem[i].high_quantity, 1, 4);
    mem::endian(&in->image.elem[i].packing, 1, 2);
    mem::endian(&in->image.elem[i].encoding, 1, 2);
    mem::endian(&in->image.elem[i].data_offset, 1, 4);
    mem::endian(&in->image.elem[i].line_padding, 1, 4);
    mem::endian(&in->image.elem[i].elem_padding, 1, 4);
  }
  
  mem::endian(in->source.offset, 2, 4);
  mem::endian(in->source.center, 2, 4);
  mem::endian(in->source.size, 2, 4);
  mem::endian(in->source.border, 4, 2);
  mem::endian(in->source.pixel_aspect, 2, 4);
  mem::endian(in->source.scan_size, 2, 4);
  
  mem::endian(&in->film.frame, 1, 4);
  mem::endian(&in->film.sequence, 1, 4);
  mem::endian(&in->film.hold, 1, 4);
  mem::endian(&in->film.frame_rate, 1, 4);
  mem::endian(&in->film.shutter, 1, 4);

  mem::endian(&in->tv.timecode, 1, 4);
  mem::endian(&in->tv.user_bits, 1, 4);
  mem::endian(in->tv.sample_rate, 2, 4);
  mem::endian(&in->tv.frame_rate, 1, 4);
  mem::endian(&in->tv.time_offset, 1, 4);
  mem::endian(&in->tv.gamma, 1, 4);
  mem::endian(&in->tv.black_level, 1, 4);
  mem::endian(&in->tv.black_gain, 1, 4);
  mem::endian(&in->tv.breakpoint, 1, 4);
  mem::endian(&in->tv.white_level, 1, 4);
  mem::endian(&in->tv.integration_times, 1, 4);
}

Str dbg(uint8_t in)
  { return str::int_to_str(in); }
  //{ return valid(&in) ? str::int_to_str(in) : "-"; }
Str dbg(uint16_t in)
  { return str::int_to_str(in); }
  //{ return valid(&in) ? str::int_to_str(in) : "-"; }
Str dbg(uint32_t in)
  { return str::int_to_str(in); }
  //{ return valid(&in) ? str::int_to_str(in) : "-"; }
  //{ return dbg_u32(in); }
Str dbg(float in)
  { return str::float_to_str(in); }
  //{ return valid(&in) ? str::float_to_str(in) : "-"; }
  //{ return dbg_u32(in); }
Str dbg(const char * in, int size)
  { return Str(in, size); }
  //{ return valid(in, size) ? Str(in, size) : "-"; }

void dbg(const Header & in)
{
  DJV_DBG_OBJ2(_DBG, "dbg");
  
  DJV_DBG_PRN("file");
  DJV_DBG_PRN("  image offset = " << dbg(in.file.image_offset));
  DJV_DBG_PRN("  version = " << dbg(in.file.version, 8));
  DJV_DBG_PRN("  file size = " << dbg(in.file.size));
  DJV_DBG_PRN("  ditto key = " << dbg(in.file.ditto_key));
  DJV_DBG_PRN("  header size = " << dbg(in.file.header_size));
  DJV_DBG_PRN("  industry header size = " << dbg(in.file.industry_header_size));
  DJV_DBG_PRN("  user header size = " << dbg(in.file.user_header_size));
  DJV_DBG_PRN("  name = " << dbg(in.file.name, 100));
  DJV_DBG_PRN("  time = " << dbg(in.file.time, 24));
  DJV_DBG_PRN("  creator = " << dbg(in.file.creator, 100));
  DJV_DBG_PRN("  project = " << dbg(in.file.project, 200));
  DJV_DBG_PRN("  copyright = " << dbg(in.file.copyright, 200));
  DJV_DBG_PRN("  encryption key = " << dbg(in.file.encryption_key));

  DJV_DBG_PRN("image");
  DJV_DBG_PRN("  orient = " << dbg(in.image.orient));
  DJV_DBG_PRN("  element size = " << dbg(in.image.elem_size));
  DJV_DBG_PRN("  size = " << dbg(in.image.size[0]) << " " <<
    dbg(in.image.size[1]));
  for (int i = 0; i < in.image.elem_size; ++i) {
    DJV_DBG_PRN("  element " << i);
    DJV_DBG_PRN("    data sign = " << dbg(in.image.elem[i].data_sign));
    DJV_DBG_PRN("    low data = " << dbg(in.image.elem[i].low_data));
    DJV_DBG_PRN("    low quantity = " << dbg(in.image.elem[i].low_quantity));
    DJV_DBG_PRN("    high data = " << dbg(in.image.elem[i].high_data));
    DJV_DBG_PRN("    high quantity = " << dbg(in.image.elem[i].high_quantity));
    DJV_DBG_PRN("    descriptor = " << dbg(in.image.elem[i].descriptor));
    DJV_DBG_PRN("    transfer = " << dbg(in.image.elem[i].transfer));
    DJV_DBG_PRN("    colorimetric = " << dbg(in.image.elem[i].colorimetric));
    DJV_DBG_PRN("    bit depth = " << dbg(in.image.elem[i].bit_depth));
    DJV_DBG_PRN("    packing = " << dbg(in.image.elem[i].packing));
    DJV_DBG_PRN("    encoding = " << dbg(in.image.elem[i].encoding));
    DJV_DBG_PRN("    data offset = " << dbg(in.image.elem[i].data_offset));
    DJV_DBG_PRN("    line padding = " << dbg(in.image.elem[i].line_padding));
    DJV_DBG_PRN("    elem padding = " << dbg(in.image.elem[i].elem_padding));
    DJV_DBG_PRN("    description = " << dbg(in.image.elem[i].description, 32));
  }
  
  DJV_DBG_PRN("source");
  DJV_DBG_PRN("  offset = " << dbg(in.source.offset[0]) << " " <<
    dbg(in.source.offset[1]));
  DJV_DBG_PRN("  center = " << dbg(in.source.center[0]) << " " <<
    dbg(in.source.center[1]));
  DJV_DBG_PRN("  size = " << dbg(in.source.size[0]) << " " <<
    dbg(in.source.size[1]));
  DJV_DBG_PRN("  file = " << dbg(in.source.file, 100));
  DJV_DBG_PRN("  time = " << dbg(in.source.time, 24));
  DJV_DBG_PRN("  input device = " << dbg(in.source.input_device, 32));
  DJV_DBG_PRN("  inputs serial = " << dbg(in.source.input_serial, 32));
  DJV_DBG_PRN("  border = " << dbg(in.source.border[0]) << " " <<
    dbg(in.source.border[1]) << " " <<  dbg(in.source.border[2]) << " " <<
    dbg(in.source.border[3]));
  DJV_DBG_PRN("  pixel aspect = " << dbg(in.source.pixel_aspect[0]) << " " <<
    dbg(in.source.pixel_aspect[1]));
  DJV_DBG_PRN("  scan size = " << dbg(in.source.scan_size[0]) << " " <<
    dbg(in.source.scan_size[1]));

  DJV_DBG_PRN("film");
  DJV_DBG_PRN("  id = " << dbg(in.film.id, 2));
  DJV_DBG_PRN("  type = " << dbg(in.film.type, 2));
  DJV_DBG_PRN("  offset = " << dbg(in.film.offset, 2));
  DJV_DBG_PRN("  prefix = " << dbg(in.film.prefix, 6));
  DJV_DBG_PRN("  count = " << dbg(in.film.count, 4));
  DJV_DBG_PRN("  format = " << dbg(in.film.format, 32));
  DJV_DBG_PRN("  frame = " << dbg(in.film.frame));
  DJV_DBG_PRN("  sequence = " << dbg(in.film.sequence));
  DJV_DBG_PRN("  hold = " << dbg(in.film.hold));
  DJV_DBG_PRN("  frame rate = " << dbg(in.film.frame_rate));
  DJV_DBG_PRN("  shutter = " << dbg(in.film.shutter));
  DJV_DBG_PRN("  frame id = " << dbg(in.film.frame_id, 32));
  DJV_DBG_PRN("  slate = " << dbg(in.film.slate, 100));

  DJV_DBG_PRN("tv");
  DJV_DBG_PRN("  timecode = " << dbg(in.tv.timecode));
  DJV_DBG_PRN("  user bits = " << dbg(in.tv.user_bits));
  DJV_DBG_PRN("  interlace = " << dbg(in.tv.interlace));
  DJV_DBG_PRN("  field = " << dbg(in.tv.field));
  DJV_DBG_PRN("  video signal = " << dbg(in.tv.video_signal));
  DJV_DBG_PRN("  sample rate = " << dbg(in.tv.sample_rate[0]) << " " <<
    dbg(in.tv.sample_rate[1]));
  DJV_DBG_PRN("  frame rate = " << dbg(in.tv.frame_rate));
  DJV_DBG_PRN("  time offset = " << dbg(in.tv.time_offset));
  DJV_DBG_PRN("  gamma = " << dbg(in.tv.gamma));
  DJV_DBG_PRN("  black level = " << dbg(in.tv.black_level));
  DJV_DBG_PRN("  black gain = " << dbg(in.tv.black_gain));
  DJV_DBG_PRN("  breakpoint = " << dbg(in.tv.breakpoint));
  DJV_DBG_PRN("  white level = " << dbg(in.tv.white_level));
  DJV_DBG_PRN("  integration times = " << dbg(in.tv.integration_times));
}

void _file(const Str & file, const image::Info & info, Header::File * out)
{
  const List<Str> & tag_label = tag::label();
  Str tag;

  out->image_offset = 2048;
  out->header_size = 2048 - 384;
  out->industry_header_size = 384;
  out->user_header_size = 0;
  out->size = 0;
  
  zero(out->version, 8);
  mem::copy("V2.0", out->version, 4);
  
  out->ditto_key = 1;  
  
  zero(out->name, 100);
  str::ascii(file, out->name, 100, false);

  zero(out->time, 24);
  tag = info.tag[tag_label[tag::TIME]];
  if (tag.size()) str::ascii(tag, out->time, 24, false);

  zero(out->creator, 100);
  tag = info.tag[tag_label[tag::CREATOR]];
  if (tag.size()) str::ascii(tag, out->creator, 100, false);

  zero(out->project, 200);
  tag = info.tag[tag_label[tag::PROJECT]];
  if (tag.size()) str::ascii(tag, out->project, 200, false);

  zero(out->copyright, 200);
  tag = info.tag[tag_label[tag::COPYRIGHT]];
  if (tag.size()) str::ascii(tag, out->copyright, 200, false);
  
  out->encryption_key = 0;
}

void _file(const Header::File & in, image::Info * info)
{
  const List<Str> & tag_label = tag::label();

  if (valid(in.time, 24))
    info->tag[tag_label[tag::TIME]] = _str(in.time, 24);
  
  if (valid(in.creator, 100))
    info->tag[tag_label[tag::CREATOR]] = _str(in.creator, 100);
  
  if (valid(in.project, 200))
    info->tag[tag_label[tag::PROJECT]] = _str(in.project, 200);
  
  if (valid(in.copyright, 200))
    info->tag[tag_label[tag::COPYRIGHT]] = _str(in.copyright, 200);
}

void _image(const image::Info & info, COLOR color, Header::Image * out)
{
  out->size[0] = info.size.x;
  out->size[1] = info.size.y;
  
  out->orient = ORIENT_LEFT_RIGHT_TOP_BOTTOM;
  
  out->elem_size = 1;
  
  switch (info.pixel) {
    case pixel::L_U8:
    case pixel::L_U16:
    case pixel::L_F16: 
    case pixel::L_F32: out->elem[0].descriptor = DESCRIPTOR_L; break;
    case pixel::RGB_U8:
    case pixel::RGB_U10:
    case pixel::RGB_U16:
    case pixel::RGB_F16: 
    case pixel::RGB_F32: out->elem[0].descriptor = DESCRIPTOR_RGB; break;
    case pixel::RGBA_U8:
    case pixel::RGBA_U16:
    case pixel::RGBA_F16: 
    case pixel::RGBA_F32: out->elem[0].descriptor = DESCRIPTOR_RGBA; break;
    default: break;
  }
  
  switch (info.pixel) {
    case pixel::RGB_U10: out->elem[0].packing = TYPE_A; break;
    default: break;
  }
  
  const int bit_depth = pixel::bit_depth(info.pixel);  
  out->elem[0].bit_depth = bit_depth;

  out->elem[0].data_sign = 0;
  out->elem[0].low_data = 0;
  switch (bit_depth) {
    case 8: out->elem[0].high_data = 255; break;
    case 10: out->elem[0].high_data = 1023; break;
    case 12: out->elem[0].high_data = 4095; break;
    case 16: out->elem[0].high_data = 65535; break;
  }
    
  if (COLOR_RAW == color) {
    out->elem[0].transfer = TRANSFER_LINEAR;
    out->elem[0].colorimetric = COLORIMETRIC_USER;
  } else {
    out->elem[0].transfer = TRANSFER_PRINT;
    out->elem[0].colorimetric = COLORIMETRIC_PRINT;
  }

  out->elem[0].encoding = 0;
  out->elem[0].data_offset = 2048;
  out->elem[0].line_padding = 0;
  out->elem[0].elem_padding = 0;
}

void _image(const Header::Image & in, image::Info * info, bool * print)
  throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "_image");
  
  info->size = Vec2i(in.size[0], in.size[1]);
  DJV_DBG_PRN("size = " << info->size);
  
  switch (in.orient) {
    case ORIENT_LEFT_RIGHT_TOP_BOTTOM: info->mirror_y = true; break;
    case ORIENT_RIGHT_LEFT_TOP_BOTTOM: info->mirror_x = true; break;
    case ORIENT_LEFT_RIGHT_BOTTOM_TOP: break;
    case ORIENT_RIGHT_LEFT_BOTTOM_TOP:
      info->mirror_x = info->mirror_y = true; break;
    case ORIENT_TOP_BOTTOM_LEFT_RIGHT:
    case ORIENT_TOP_BOTTOM_RIGHT_LEFT:
    case ORIENT_BOTTOM_TOP_LEFT_RIGHT:
    case ORIENT_BOTTOM_TOP_RIGHT_LEFT: break;
  }

  if (in.elem_size != 1)
    throw Error(plugin_name, error_unsupported);

  int pixel = -1;
  switch (in.elem[0].packing) {
    case PACK: {
      int channels = 0;
      switch (in.elem[0].descriptor) {
        case DESCRIPTOR_L: channels = 1; break;
        case DESCRIPTOR_RGB: channels = 3; break;
        case DESCRIPTOR_RGBA: channels = 4; break;
      }
      pixel = pixel::pixel(channels, in.elem[0].bit_depth);
    } break;
    case TYPE_A:
      switch (in.elem[0].descriptor) {
        case DESCRIPTOR_RGB:
          switch (in.elem[0].bit_depth) {
            case 10: pixel = pixel::RGB_U10; break;
          }
          break;
      }
      break;
  }
  if (-1 == pixel)
    throw Error(plugin_name, error_unsupported);
  info->pixel = pixel::PIXEL(pixel);
  DJV_DBG_PRN("pixel = " << info->pixel);
  
  switch (pixel::bit_depth(info->pixel)) {
    case 8:
    case 10: info->align = 4; break;
  }

  if (in.elem[0].encoding)
    throw Error(plugin_name, error_unsupported);
  
  if (valid(&in.elem[0].line_padding) && in.elem[0].line_padding)
    throw Error(plugin_name, error_unsupported);
  
  if (print)
    *print = TRANSFER_PRINT == in.elem[0].transfer;
}

void _source(const image::Info & info, Header::Source * out)
{
  const List<Str> & tag_label = tag::label();
  const List<Str> & dpx_tag_label = dpx::tag_label();
  Str tag;
  
  tag = info.tag[dpx_tag_label[TAG_SOURCE_OFFSET]];
  if (tag.size()) {
    const List<Str> tmp = str::split(tag, " ");
    if (2 == tmp.size()) {
      out->offset[0] = str::str_to_int(tmp[0]);
      out->offset[1] = str::str_to_int(tmp[1]);
    }
  }

  tag = info.tag[dpx_tag_label[TAG_SOURCE_CENTER]];
  if (tag.size()) {
    const List<Str> tmp = str::split(tag, " ");
    if (2 == tmp.size()) {
      out->center[0] = str::str_to_float(tmp[0]);
      out->center[1] = str::str_to_float(tmp[1]);
    }
  }

  tag = info.tag[dpx_tag_label[TAG_SOURCE_SIZE]];
  if (tag.size()) {
    const List<Str> tmp = str::split(tag, " ");
    if (2 == tmp.size()) {
      out->size[0] = str::str_to_int(tmp[0]);
      out->size[1] = str::str_to_int(tmp[1]);
    }
  }
  
  zero(out->file, 100);
  tag = info.tag[dpx_tag_label[TAG_SOURCE_FILE]];
  if (tag.size()) str::ascii(tag, out->file, 100, false);

  zero(out->time, 24);
  tag = info.tag[tag_label[TAG_SOURCE_TIME]];
  if (tag.size()) str::ascii(tag, out->time, 24, false);

  zero(out->input_device, 32);
  tag = info.tag[dpx_tag_label[TAG_SOURCE_INPUT_DEVICE]];
  if (tag.size()) str::ascii(tag, out->input_device, 32, false);

  zero(out->input_serial, 32);
  tag = info.tag[dpx_tag_label[TAG_SOURCE_INPUT_SERIAL]];
  if (tag.size()) str::ascii(tag, out->input_serial, 32, false);

  tag = info.tag[dpx_tag_label[TAG_SOURCE_BORDER]];
  if (tag.size()) {
    const List<Str> tmp = str::split(tag, " ");
    if (4 == tmp.size()) {
      out->border[0] = str::str_to_int(tmp[0]);
      out->border[1] = str::str_to_int(tmp[1]);
      out->border[2] = str::str_to_int(tmp[2]);
      out->border[3] = str::str_to_int(tmp[3]);
    }
  }

  tag = info.tag[dpx_tag_label[TAG_SOURCE_PIXEL_ASPECT]];
  if (tag.size()) {
    const List<Str> tmp = str::split(tag, " ");
    if (2 == tmp.size()) {
      out->pixel_aspect[0] = str::str_to_int(tmp[0]);
      out->pixel_aspect[1] = str::str_to_int(tmp[1]);
    }
  }

  tag = info.tag[dpx_tag_label[TAG_SOURCE_SCAN_SIZE]];
  if (tag.size()) {
    const List<Str> tmp = str::split(tag, " ");
    if (2 == tmp.size()) {
      out->scan_size[0] = str::str_to_float(tmp[0]);
      out->scan_size[1] = str::str_to_float(tmp[1]);
    }
  }
}

void _source(const Header::Source & in, image::Info * info)
{
  const List<Str> & tag_label = tag::label();
  const List<Str> & dpx_tag_label = dpx::tag_label();
  
  if (valid(&in.offset[0]) && valid(&in.offset[1]))
    info->tag[dpx_tag_label[TAG_SOURCE_OFFSET]] = str::join(List<Str>() <<
      str::int_to_str(in.offset[0]) << str::int_to_str(in.offset[1]), " ");
  
  if (valid(&in.center[0]) && valid(&in.center[1]))
    info->tag[dpx_tag_label[TAG_SOURCE_CENTER]] = str::join(List<Str>() <<
      str::float_to_str(in.center[0]) << str::float_to_str(in.center[1]), " ");
  
  if (valid(&in.size[0]) && valid(&in.size[1]))
    info->tag[dpx_tag_label[TAG_SOURCE_SIZE]] = str::join(List<Str>() <<
      str::int_to_str(in.size[0]) << str::int_to_str(in.size[1]), " ");
  
  if (valid(in.file, 100))
    info->tag[dpx_tag_label[TAG_SOURCE_FILE]] = _str(in.file, 100);
  
  if (valid(in.time, 24))
    info->tag[tag_label[TAG_SOURCE_TIME]] = _str(in.time, 24);
  
  if (valid(in.input_device, 32))
    info->tag[dpx_tag_label[TAG_SOURCE_INPUT_DEVICE]] =
      _str(in.input_device, 32);
  
  if (valid(in.input_serial, 32))
    info->tag[dpx_tag_label[TAG_SOURCE_INPUT_SERIAL]] =
      _str(in.input_serial, 32);
  
  if (valid(&in.border[0]) && valid(&in.border[1]) &&
    valid(&in.border[2]) && valid(&in.border[3])
  )
    info->tag[dpx_tag_label[TAG_SOURCE_BORDER]] = str::join(List<Str>() <<
      str::int_to_str(in.border[0]) << str::int_to_str(in.border[1]) <<
      str::int_to_str(in.border[2]) << str::int_to_str(in.border[3]), " ");
  
  if (valid(&in.pixel_aspect[0]) && valid(&in.pixel_aspect[1]))
    info->tag[dpx_tag_label[TAG_SOURCE_PIXEL_ASPECT]] = str::join(
      List<Str>() << str::int_to_str(in.pixel_aspect[0]) <<
      str::int_to_str(in.pixel_aspect[1]), " ");
  
  if (valid(&in.scan_size[0]) && valid(&in.scan_size[1]))
    info->tag[dpx_tag_label[TAG_SOURCE_SCAN_SIZE]] =
      str::join(List<Str>() <<
      str::float_to_str(in.scan_size[0]) <<
      str::float_to_str(in.scan_size[1]), " ");
}

void _film(const image::Info & info, Header::Film * out)
{
  const List<Str> & tag_label = tag::label();
  const List<Str> & dpx_tag_label = dpx::tag_label();
  Str tag;

  zero(out->id, 2);
  zero(out->type, 2);
  zero(out->offset, 2);
  zero(out->prefix, 6);
  zero(out->count, 4);
  tag = info.tag[tag_label[tag::KEYCODE]];
  if (tag.size()) {
    int id = 0, type = 0, prefix = 0, count = 0, offset = 0;
    time::str_to_keycode(tag, &id, &type, &prefix, &count, &offset);
    str::ascii(str::int_to_str(id), out->id, 2, false);
    str::ascii(str::int_to_str(type), out->type, 2, false);
    str::ascii(str::int_to_str(offset), out->offset, 2, false);
    str::ascii(str::int_to_str(prefix), out->prefix, 6, false);
    str::ascii(str::int_to_str(count), out->count, 4, false);
  }
  
  zero(out->format, 32);
  tag = info.tag[dpx_tag_label[TAG_FILM_FORMAT]];
  if (tag.size()) str::ascii(tag, out->format, 32, false);

  tag = info.tag[dpx_tag_label[TAG_FILM_FRAME]];
  if (tag.size()) out->frame = str::str_to_int(tag);

  tag = info.tag[dpx_tag_label[TAG_FILM_SEQUENCE]];
  if (tag.size()) out->sequence = str::str_to_int(tag);

  tag = info.tag[dpx_tag_label[TAG_FILM_HOLD]];
  if (tag.size()) out->hold = str::str_to_int(tag);

  tag = info.tag[dpx_tag_label[TAG_FILM_FRAME_RATE]];
  if (tag.size()) out->frame_rate = str::str_to_float(tag);

  tag = info.tag[dpx_tag_label[TAG_FILM_SHUTTER]];
  if (tag.size()) out->shutter = str::str_to_float(tag);

  zero(out->frame_id, 32);
  tag = info.tag[dpx_tag_label[TAG_FILM_FRAME_ID]];
  if (tag.size()) str::ascii(tag, out->frame_id, 32, false);

  zero(out->slate, 100);
  tag = info.tag[dpx_tag_label[TAG_FILM_SLATE]];
  if (tag.size()) str::ascii(tag, out->slate, 100, false);
}

void _film(const Header::Film & in, Info * info)
{
  const List<Str> & tag_label = tag::label();
  const List<Str> & dpx_tag_label = dpx::tag_label();
  
  if (valid(in.id, 2) && valid(in.type, 2) &&
    valid(in.offset, 2) && valid(in.prefix, 6) &&
    valid(in.count, 4)
  )
    info->tag[tag_label[tag::KEYCODE]] = time::keycode_to_str(
      str::str_to_int(_str(in.id, 2)), str::str_to_int(_str(in.type, 2)),
      str::str_to_int(_str(in.prefix, 6)), str::str_to_int(_str(in.count, 4)),
      str::str_to_int(_str(in.offset, 2)));

  if (valid(in.format, 32))
    info->tag[dpx_tag_label[TAG_FILM_FORMAT]] = _str(in.format, 32);
  
  if (valid(&in.frame))
    info->tag[dpx_tag_label[TAG_FILM_FRAME]] = str::int_to_str(in.frame);
  
  if (valid(&in.sequence))
    info->tag[dpx_tag_label[TAG_FILM_SEQUENCE]] =
      str::int_to_str(in.sequence);
  
  if (valid(&in.hold))
    info->tag[dpx_tag_label[TAG_FILM_HOLD]]  = str::int_to_str(in.hold);
  
  if (valid(&in.frame_rate) && in.frame_rate > 0.000001) {
    info->time.speed = time::float_to_speed(in.frame_rate);
    info->tag[dpx_tag_label[TAG_FILM_FRAME_RATE]] =
      str::float_to_str(in.frame_rate);
  }
  
  if (valid(&in.shutter))
    info->tag[dpx_tag_label[TAG_FILM_SHUTTER]] =
      str::float_to_str(in.shutter);
  
  if (valid(in.frame_id, 32))
    info->tag[dpx_tag_label[TAG_FILM_FRAME_ID]] = _str(in.frame_id, 32);
  
  if (valid(in.slate, 100))
    info->tag[dpx_tag_label[TAG_FILM_SLATE]] = _str(in.slate, 100);
}

void _tv(const image::Info & info, Header::Tv * out)
{
  const List<Str> & tag_label = tag::label();
  const List<Str> & dpx_tag_label = dpx::tag_label();
  Str tag;

  tag = info.tag[tag_label[tag::TIMECODE]];
  if (tag.size()) out->timecode = time::str_to_timecode(tag);

  tag = info.tag[dpx_tag_label[TAG_TV_INTERLACE]];
  if (tag.size()) out->interlace = str::str_to_int(tag);

  tag = info.tag[dpx_tag_label[TAG_TV_FIELD]];
  if (tag.size()) out->field = str::str_to_int(tag);

  tag = info.tag[dpx_tag_label[TAG_TV_VIDEO_SIGNAL]];
  if (tag.size()) out->video_signal = str::str_to_int(tag);
  
  tag = info.tag[dpx_tag_label[TAG_TV_SAMPLE_RATE]];
  if (tag.size()) {
    const List<Str> tmp = str::split(tag, " ");
    if (2 == tmp.size()) {
      out->sample_rate[0] = str::str_to_float(tmp[0]);
      out->sample_rate[1] = str::str_to_float(tmp[1]);
    }
  }
  
  tag = info.tag[dpx_tag_label[TAG_TV_FRAME_RATE]];
  if (tag.size()) out->frame_rate = str::str_to_float(tag);

  tag = info.tag[dpx_tag_label[TAG_TV_TIME_OFFSET]];
  if (tag.size()) out->time_offset = str::str_to_float(tag);

  tag = info.tag[dpx_tag_label[TAG_TV_GAMMA]];
  if (tag.size()) out->gamma = str::str_to_float(tag);

  tag = info.tag[dpx_tag_label[TAG_TV_BLACK_LEVEL]];
  if (tag.size()) out->black_level = str::str_to_float(tag);

  tag = info.tag[dpx_tag_label[TAG_TV_BLACK_GAIN]];
  if (tag.size()) out->black_gain = str::str_to_float(tag);

  tag = info.tag[dpx_tag_label[TAG_TV_BREAK_POINT]];
  if (tag.size()) out->breakpoint = str::str_to_float(tag);

  tag = info.tag[dpx_tag_label[TAG_TV_WHITE_LEVEL]];
  if (tag.size()) out->white_level = str::str_to_float(tag);

  tag = info.tag[dpx_tag_label[TAG_TV_INTEGRATION_TIMES]];
  if (tag.size()) out->integration_times = str::str_to_float(tag);
}

void _tv(const Header::Tv & in, Info * info)
{
  const List<Str> & tag_label = tag::label();
  const List<Str> & dpx_tag_label = dpx::tag_label();
  
  if (valid(&in.timecode))
    info->tag[tag_label[tag::TIMECODE]] =
      time::timecode_to_str(in.timecode);
  
  if (valid(&in.interlace))
    info->tag[dpx_tag_label[TAG_TV_INTERLACE]] =
      str::int_to_str(in.interlace);
  
  if (valid(&in.field))
    info->tag[dpx_tag_label[TAG_TV_FIELD]] = str::int_to_str(in.field);
  
  if (valid(&in.video_signal))
    info->tag[dpx_tag_label[TAG_TV_VIDEO_SIGNAL]] =
      str::int_to_str(in.video_signal);
  
  if (valid(&in.sample_rate[0]) && valid(&in.sample_rate[1]))
    info->tag[dpx_tag_label[TAG_TV_SAMPLE_RATE]] = str::join(List<Str>() <<
      str::float_to_str(in.sample_rate[0]) <<
      str::float_to_str(in.sample_rate[1]), " ");
  
  if (valid(&in.frame_rate) && in.frame_rate > 0.000001) {
    info->time.speed = time::float_to_speed(in.frame_rate);
    info->tag[dpx_tag_label[TAG_TV_FRAME_RATE]] =
      str::float_to_str(in.frame_rate);
  }
  
  if (valid(&in.time_offset))
    info->tag[dpx_tag_label[TAG_TV_TIME_OFFSET]] =
      str::float_to_str(in.time_offset);
  
  if (valid(&in.gamma))
    info->tag[dpx_tag_label[TAG_TV_GAMMA]] = str::float_to_str(in.gamma);
  
  if (valid(&in.black_level))
    info->tag[dpx_tag_label[TAG_TV_BLACK_LEVEL]] =
      str::float_to_str(in.black_level);
  
  if (valid(&in.black_gain))
    info->tag[dpx_tag_label[TAG_TV_BLACK_GAIN]] =
      str::float_to_str(in.black_gain);
  
  if (valid(&in.breakpoint))
    info->tag[dpx_tag_label[TAG_TV_BREAK_POINT]] =
      str::float_to_str(in.breakpoint);
  
  if (valid(&in.white_level))
    info->tag[dpx_tag_label[TAG_TV_WHITE_LEVEL]] =
      str::float_to_str(in.white_level);
  
  if (valid(&in.integration_times))
    info->tag[dpx_tag_label[TAG_TV_INTEGRATION_TIMES]] =
      str::float_to_str(in.integration_times);
}

}

void header_set(djv::Io * io, const Str & file, const image::Info & info,
  ENDIAN endian, COLOR color
) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "header_set");
  DJV_DBG_PRN("info = " << info);
  DJV_DBG_PRN("color = " << color);
  
  Header header;
  mem::set(0xff, &header, sizeof(Header));
  _file(file, info, &header.file);
  _image(info, color, &header.image);
  _source(info, &header.source);
  _film(info, &header.film);
  _tv(info, &header.tv);
  dbg(header);
  
  mem::ENDIAN file_endian = mem::endian_;
  if (MSB == endian) file_endian = mem::MSB;
  else if (LSB == endian) file_endian = mem::LSB;
  if (file_endian != mem::endian()) {
    DJV_DBG_PRN("endian");
    io->endian(true);
    _endian(&header);
  }
  mem::copy(mem::MSB == file_endian ? magic[0] : magic[1],
    &header.file.magic, 4);
  io->set(&header, sizeof(Header));
}

void header_update(djv::Io * io) throw (Error)
{
  const size_t size = io->position();
  io->position(12);
  io->set_u32(size);
}

void header_get(djv::Io * io, Info * info, bool * print) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "header_get");
  
  Header header;
  io->get(&header, sizeof(Header));
  DJV_DBG_PRN("magic = " << Str((char *)&header.file.magic, 4));
  if (0 == mem::compare(&header.file.magic, magic[0], 4))
    info->endian = mem::MSB;
  else if (0 == mem::compare(&header.file.magic, magic[1], 4))
    info->endian = mem::LSB;
  else
    throw Error(plugin_name, error_unrecognized);
  if (info->endian != mem::endian()) {
    DJV_DBG_PRN("endian");
    io->endian(true);
    _endian(&header);
  }
  dbg(header);
  
  _file(header.file, info);
  _image(header.image, info, print);
  _source(header.source, info);
  _film(header.film, info);
  _tv(header.tv, info);

  if (header.file.image_offset)
    io->position(header.file.image_offset);
}

}
}
}
}

