/*------------------------------------------------------------------------------
 lib/djv_ui/file_browser_widget.cxx
 
 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_ui/file_browser_widget.cxx"

#include "_file_browser.h"
#include <djv_base/user.h>
#include <djv_image/time.h>
#include <djv_ui/style.h>
#include <djv_ui/icon.h>
#include <FL/Fl.H>
#include <FL/fl_draw.H>

namespace djv {
namespace ui {
namespace file_browser {

Browser::Browser() :
  Fl_Browser_(0, 0, 0, 0),
  _text_height(0), _icon_size(0), _thumbnail_size(0), _margin(0),
  _images(false)
{
  DJV_DBG_OBJ2(_DBG, "Browser::Browser");
  
  end();

  Fl_Widget::type(FL_HOLD_BROWSER);
  has_scrollbar(VERTICAL_ALWAYS);
  box(FL_FLAT_BOX);
  //Fl_Browser_::scrollbar.slider(FL_THIN_UP_BOX);
  //Fl_Browser_::hscrollbar.slider(FL_THIN_UP_BOX);

  for (uint i = 0; i < Dialog::sort_size; ++i)
    _column_list[i] = 0;

  callback(widget_callback, this);
}

Browser::~Browser()
{
  DJV_DBG_OBJ2(_DBG, "Browser::~Browser");
  
  del();
}

void Browser::set(const List<File> & in, bool images)
{
  using file_browser::Item;
  
  DJV_DBG_OBJ2(_DBG, "Browser::set");
  DJV_DBG_PRN("in = " << in.size());
  DJV_DBG_PRN("images = " << images);
    
  del();
  
  const int size = in.size();
  _item_list.resize(size);
  for (int i = 0; i < size; ++i)
    _item_list[i].set(in[i], i % 2);
  for (int i = 1; i < size; ++i)
    _item_list[i].prev = &_item_list[i - 1];
  for (int i = 0; i < size - 1; ++i)
    _item_list[i].next = &_item_list[i + 1];
  
  _images = images;
  
  deselect();
  new_list();
}

File Browser::get() const
{
  using file_browser::Item;
  
  const Item * item = reinterpret_cast<Item *>(selection());
  return item ? item->get() : File();
}

void Browser::del()
{
  DJV_DBG_OBJ2(_DBG, "Browser::del");
  
  _item_list.clear();
  _images_work_list.clear();
}

void Browser::column_list(const List<int> & in)
{
  DJV_DBG_OBJ2(_DBG, "Browser::column_list");
  DJV_DBG_PRN("in = " << in);
  
  for (uint i = 0; i < Dialog::sort_size; ++i)
    _column_list[i] = i < in.size() ? in[i] : 0;
  
  redraw();
}

int Browser::handle(int in)
{
  using file_browser::Item;
  
  switch (in) {
  
    case FL_KEYBOARD:
      switch (Fl::event_key()) { 
        case FL_Enter:
        case FL_KP_Enter: {
          Item * item = reinterpret_cast<Item *>(selection());
          if (item) action_sig.emit(item->get());
          return 1;
        }
      }
      break;
  }
  return Fl_Browser_::handle(in);
}

void Browser::draw()
{
  Fl_Browser_::draw();
  widget::Base::draw();
}

int Browser::item_height(void * in) const
{
  using file_browser::Item;
  
  Item * item = static_cast<Item *>(in);
  
  int height = _icon_size;
  if (_images && item)
    if (item->image_valid())
      height = math::max(math::max(_thumbnail_size, _text_height * 3), height);

  return math::max(_text_height, height) + _margin * 2;
}
  
void Browser::item_draw(void * in, int x, int y, int w, int h) const
{
  using file_browser::Item;
  
  Item * item = static_cast<Item *>(in);
  if (! item) return;

  DJV_DBG_OBJ2(_DBG, "Browser::item_draw");
  DJV_DBG_PRN("in = " << item->get());
  
  fl_font(style::style()->font(), style::style()->font_size());
  const File & file = item->get();
  const int margin = style::style()->margin_widget();
  const int text_off = fl_height() - fl_descent();
  int _x, _y;
  
  // Background.
  
  if (item != selection()) {
    const Fl_Color color_list [] = { FL_BACKGROUND_COLOR,
      FL_BACKGROUND2_COLOR };
    fl_color(color_list[item->odd()]);
    fl_rectf(x, y, w, h);
  }
  
  // Name, icon, and thumbnail image.
  
  fl_color(in == selection() ? FL_BLACK : FL_FOREGROUND_COLOR);
  fl_push_clip(x, y, _column_list[0], h);

  _x = x + margin;
  _y = y + margin;
  
  Fl_Image * image = 0;
  if (_images) {
    if (! item->image_init()) {
      Browser * that = const_cast<Browser *>(this);
      that->_images_work_list.push_back(item);
      Fl::remove_idle(image_callback, that);
      Fl::add_idle(image_callback, that);
    } else {
      image = item->image(_thumbnail_size);
      if (image) image->draw(_x, y + h / 2 - image->h() / 2);
    }
  }
  
  if (! image) {
    Fl_Image * icon = 0;
    switch (file.type) {
      case djv::file::SEQ: icon = ui::icon::get("file_seq"); break;
      case djv::file::DIR: icon = ui::icon::get("dir"); break;
      default: icon = ui::icon::get("file"); break;
    }
    if (icon) icon->draw(_x, y + h / 2 - icon->h() / 2);
  }
  
  _x += (_images ? math::max(style::style()->thumbnail_size(),
    style::style()->icon_size()) : style::style()->icon_size()) + margin;
  
  if (image) {
    const djv::image::io::Info & io_info = item->image_io_info();
  
    _y = y + (h / 2 - fl_height() * 3 / 2);
    
    Str tmp = file.get(-1, false);
    fl_draw(tmp.c_str(), _x, _y + text_off);
    
    tmp = str::Format("%1x%2:%3 %4").
      arg(io_info.size.x).arg(io_info.size.y).arg(aspect(io_info.size), 0, -2).
      arg(djv::image::pixel::pixel_label()[io_info.pixel]);
    fl_draw(tmp.c_str(), _x, _y + fl_height() + text_off);
    
    if (style::style()->timecode()) {
      tmp = str::Format("%1@%2").
        arg(djv::image::time::timecode_to_str(
          djv::image::time::frame_to_timecode(
          io_info.time.end - io_info.time.start + 1, io_info.time.speed))).
        arg(djv::image::time::speed_to_float(io_info.time.speed));
    } else {
      tmp = str::Format("%1@%2").
        arg(io_info.time.end - io_info.time.start + 1).
        arg(djv::image::time::speed_to_float(io_info.time.speed));
    }
    fl_draw(tmp.c_str(), _x, _y + fl_height() * 2 + text_off);
    
  } else {
  
    _y = y + (h / 2 - fl_height() / 2);
    
    fl_draw(file.get(-1, false).c_str(), _x, _y + text_off);
  }
  
  fl_pop_clip();

  // Information.
  
  int column = 0;

  _x = x + _column_list[column];
  _y = y + (h / 2 - fl_height() / 2);
  ++column;
  
  fl_push_clip(_x, y, _column_list[column], h);
  Str tmp = djv::file::size_label(file.size);
  fl_draw(
    tmp.c_str(),
    _x + _column_list[column] - margin - str_width(tmp),
    _y + text_off
  );
  fl_pop_clip();
  _x += _column_list[column];
  ++column;
  
#if ! defined(WIN32)

  fl_push_clip(_x, y, _column_list[column], h);
  tmp = user::uid_to_str(file.user);
  fl_draw(
    tmp.c_str(),
    _x + _column_list[column] - margin - str_width(tmp),
    _y + text_off
  );
  fl_pop_clip();
  _x += _column_list[column];
  ++column;

#endif

  fl_push_clip(_x, y, _column_list[column], h);
  tmp = djv::time::time_to_str(file.time);
  fl_draw(
    tmp.c_str(),
    _x + _column_list[column] - margin - (str_width(tmp) +
    Fl_Browser_::scrollbar.w()),
    _y + text_off
  );
  fl_pop_clip();
  ++column;
  
  // Layout.
  
  _x = x + _column_list[0];
  for (uint i = 1; i < Dialog::sort_size; ++i) {
    if (item != selection()) {
      fl_color(fl_darker(color()));
      fl_line(_x - 1, y, _x - 1, y + h - 1);
      fl_color(fl_lighter(color()));
      fl_line(_x, y, _x, y + h - 1);
    } else {
      fl_color(FL_BLACK);
      fl_line(_x - 1, y, _x - 1, y + h - 1);
    }
    _x += _column_list[i];
  }
}

void Browser::dirty()
{
  labelfont(style::style()->font());
  labelsize(style::style()->font_size());

  fl_font(style::style()->font(), style::style()->font_size());
  _text_height = fl_height();
  _icon_size = style::style()->icon_size();
  _thumbnail_size = style::style()->thumbnail_size();
  _margin = style::style()->margin_widget();

  widget::Base::dirty();
}

void Browser::resize(int x, int y, int w, int h)
{
  Fl_Browser_::resize(x, y, w, h);
  widget::Base::resize(x, y, w, h);
}

void Browser::widget_callback()
{
  using file_browser::Item;
  
  if (Fl::event_clicks()) {

    Fl::event_clicks(0);
    Item * item = reinterpret_cast<Item *>(selection());
    if (item) action_sig.emit(item->get());

  } else {

    sig.emit(get());
  }
}

void Browser::image_callback()
{
  using file_browser::Item;
  
  DJV_DBG_OBJ2(_DBG, "Browser::image_callback");
  DJV_DBG_PRN("images = " << _images_work_list.size());
  DJV_DBG_PRN("visible = " << visible_r());
  
  if (_images_work_list.size()) {
    Item * item = _images_work_list.front();
    _images_work_list.pop_front();
    if (displayed(item)) {
      item->image(_thumbnail_size);
      redraw_line(item);
    }
  }
  
  if (! _images_work_list.size())
    Fl::remove_idle(image_callback, this);
  
  if (! visible_r()) {
    Fl::remove_idle(image_callback, this);
    _images_work_list.clear();
  }
}

}
}
}
