/*------------------------------------------------------------------------------
 lib/djv_ui/image.cxx
 
 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_ui/image.cxx"

#include "image.h"
#include <djv_image/image.h>
#include <djv_ui/app.h>

namespace djv {
namespace ui {
namespace image {

Widget::Widget() :
  _data(0),
  _zoom(1.0),
  _mirror_x(false), _mirror_y(false),
  _scale(SCALE(0)),
  _rotate(ROTATE(0)),
  _channel(CHANNEL(0)),
  _filter(FILTER(0))
{}

void Widget::set(const djv::image::gl::data::Data * in)
{
  _data = in;
}

void Widget::view(const Vec2i & in)
{
  if (in == _view) return;
  _view = in;
  sig.emit(true);
}

void Widget::zoom(double in)
{
  if (in == _zoom) return;
  
  DJV_DBG_OBJ2(_DBG, "Widget::zoom");
  DJV_DBG_PRN("in = " << in);
  
  _zoom = in;
  sig.emit(true);
}

void Widget::zoom(double in, const Vec2i & focus)
{
  if (in == _zoom) return;
  
  DJV_DBG_OBJ2(_DBG, "Widget::zoom");
  DJV_DBG_PRN("in = " << in);
  DJV_DBG_PRN("focus = " << focus);
  
  const double scale = in / _zoom;
  _view = Vec2i(
    focus.x + math::floor((_view.x - focus.x) * scale),
    focus.y + math::floor((_view.y - focus.y) * scale));
  _zoom = in;
  sig.emit(true);
}

void Widget::zoom_in(const Vec2i & in) { zoom(_zoom * 2.0, in); }
void Widget::zoom_in() { zoom_in(geom().size / 2); }
void Widget::zoom_out(const Vec2i & in) { zoom(_zoom / 2.0, in); }
void Widget::zoom_out() { zoom_out(geom().size / 2); }
void Widget::zoom_reset() { zoom(1.0, geom().size / 2); }

void Widget::view_zero() { _view = Vec2i(); _zoom = 1.0; }
void Widget::view_center() { view(geom().size / 2 - screen_size() / 2); }

void Widget::view_fit()
{
  if (! _data) return;
  if (! _data->get()) return;
  
  DJV_DBG_OBJ2(_DBG, "Widget::view_fit");
  DJV_DBG_PRN("widget size = " << Vec2i(w(), h()));
  
  Vec2i size =
    image::size(_data->get()->size(), _scale, _rotate) *
    djv::image::data::proxy_scale(_data->get()->proxy());
  DJV_DBG_PRN("data size = " << size);
  
  _zoom = math::min(geom().w / double(size.x), geom().h / double(size.y));
  _view = vec::to_int(
    vec::to_float(geom().size) / 2.0 - vec::to_float(size) * _zoom / 2.0);
  DJV_DBG_PRN("view = " << _view);
  DJV_DBG_PRN("zoom = " << _zoom);
  
  sig.emit(true);
}

void Widget::mirror(bool x, bool y)
{
  _mirror_x = x;
  _mirror_y = y;
}

void Widget::scale(SCALE in) { _scale = in; }
void Widget::rotate(ROTATE in) { _rotate = in; }
void Widget::profile(const Profile & in) { _profile = in; }
void Widget::display(const Display & in) { _display = in; }
void Widget::channel(CHANNEL in) { _channel = in; }
void Widget::filter(FILTER in) { _filter = in; }

void Widget::background(const djv::image::Color & in) { _background = in; }

void Widget::draw()
{
  gl::Widget::draw();

  if (! visible()) return;
    
  DJV_DBG_OBJ2(_DBG, "Widget::draw");
  DJV_DBG_PRN("size = " << Vec2i(w(), h()));
  DJV_DBG_PRN("view = " << _view);
  DJV_DBG_PRN("zoom = " << _zoom);

  //if (! valid()) {
    const Box2i geom = this->geom();
    djv::image::gl::ortho(geom.w, geom.h);
    DJV_DBG_GL(glViewport(0, 0, geom.w, geom.h));
  //}
  
  //DJV_DBG_GL(glClearColor(0, 1, 0, 0));
  djv::image::Color background(djv::image::pixel::RGB_F32);
  djv::image::color::convert(_background, &background);
  DJV_DBG_GL(glClearColor(
    background.get_f32(0), background.get_f32(1), background.get_f32(2), 0.0));
  DJV_DBG_GL(glClear(GL_COLOR_BUFFER_BIT));
  
  if (! _data) return;
  if (! _data->get()) return;
  const djv::image::data::Info & info = _data->get()->info();
  DJV_DBG_PRN("info = " << info);

  djv::image::gl::data::Options options;
  options.mirror_x = _mirror_x;
  options.mirror_y = _mirror_y;
  options.scale =
    vec::ceil(vec::to_float(image::size(info.size, _scale)) * _zoom);
  options.rotate = image::rotate(_rotate);
  options.profile = _profile;
  options.display = _display;
  options.channel = _channel;
  options.filter = _filter;
  
  // XXX
  
  const int proxy_scale = djv::image::data::proxy_scale(info.proxy);
  DJV_DBG_PRN("proxy scale = " << proxy_scale);
  options.position = image::position(_view, options.scale * proxy_scale,
    _rotate);
  options.scale *= proxy_scale;
  
  try { _data->draw(options); }
  catch (Error in) { app::app()->error(in); }
}

Vec2i Widget::screen_size() const
{
  if (! _data) return Vec2i();
  if (! _data->get()) return Vec2i();
  
  DJV_DBG_OBJ2(_DBG, "Widget::screen_size");
  DJV_DBG_PRN("zoom = " << _zoom);
  
  const djv::image::data::Info & info = _data->get()->info();
  DJV_DBG_PRN("info = " << info);
  
  const Vec2i size = image::size(info.size, _scale, _rotate);
  DJV_DBG_PRN("size = " << size);
  
  const Vec2i out = vec::ceil(
    vec::to_float(size * djv::image::data::proxy_scale(info.proxy)) * _zoom);
  DJV_DBG_PRN("out = " << out);
  return out;
}

}
}
}
