/*------------------------------------------------------------------------------
 lib/djv_ui/float.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#include "num_float.h"
#include <djv_ui/layout_row.h>
#include <djv_ui/button.h>
#include <djv_ui/text.h>
#include <djv_ui/slider.h>
#include <FL/Fl.H>

namespace djv {
namespace ui {
namespace num {

Float_Edit::Float_Edit()
{
  // Widgets.
  
  _widget = new text::Edit;
  _widget->size_str("000.00");
  
  Inc * inc_widget = new Inc;
  
  // Layout.
  
  size_policy(FIXED, FIXED);
  align_policy(LEFT, CENTER);
  
  layout::Horizontal * layout = new layout::Horizontal(this);
  layout->spacing(2);
  layout->margin(0);
  layout->add(_widget);
  layout->stretch(_widget);
  layout->add(inc_widget);

  // Initialize.
  
  value_update();

  // Callbacks.
  
  _widget->sig.set(this, widget_callback);
  inc_widget->inc_sig.set(this, inc_callback);
  inc_widget->dec_sig.set(this, dec_callback);
}

int Float_Edit::handle(int in)
{
  if (Float_Base::handle(in)) return 1;
  return widget::Widget::handle(in);
}

layout::Item * Float_Edit::group_align() { return _widget->group_align(); }

void Float_Edit::widget_callback(const Str & in)
{
  set(str::str_to_float(in));
  value_update();
}

void Float_Edit::inc_callback(bool) { inc_small_action(); }
void Float_Edit::dec_callback(bool) { dec_small_action(); }

void Float_Edit::value_update()
{
  callbacks(false);
  
  _widget->set(str::float_to_str(get(), precision()));
  
  callbacks(true);
}

Float_Display::Float_Display() :
  _precision(0)
{
  // Widgets.
  
  _widget = new text::Display;
  _widget->size_str("000.00");
  
  // Layout.
  
  size_policy(FIXED, FIXED);
  align_policy(LEFT, CENTER);
  
  layout::Horizontal * layout = new layout::Horizontal(this);
  layout->spacing(0);
  layout->margin(0);
  layout->add(_widget);
}

void Float_Display::set(double in)
{
  _widget->set(str::float_to_str(in, _precision));
}

void Float_Display::precision(int in)
{
  _precision = in;
}

void Float_Display::text_color(const Fl_Color & in)
{
  widget::Widget::text_color(in);
  _widget->text_color(in);
}

void Float_Display::dirty()
{
  widget::Widget::dirty();
  _widget->text_color(text_color());
}

Float_Slider::Float_Slider() :
  _label(0)
{
  clamp(false);
  
  // Widgets.
  
  _widget = new text::Edit;
  _widget->size_str("000.00");
  
  Inc * inc_widget = new Inc;

  _slider = new slider::Float;
  
  _default_widget = new button::Tool("tool_x");
  _default_widget->tooltip("Reset value.");
  _default_widget->deactivate();
  _default_widget->hide();
  
  // Layout.
  
  size_policy(STRETCH, FIXED);
  
  _layout = new layout::Horizontal(this);
  _layout->spacing(2);
  _layout->margin(0);
  _layout->add(_widget);
  _layout->add(inc_widget);
  _layout->add(_slider);
  _layout->stretch(_slider);
  _layout->add(_default_widget);

  // Initialize.
  
  range(0.0, 1.0);
  
  range_update();
  value_update();

  // Callbacks.
  
  _widget->sig.set(this, widget_callback);
  inc_widget->inc_sig.set(this, inc_callback);
  inc_widget->dec_sig.set(this, dec_callback);
  _slider->sig.set(this, slider_callback);
  _default_widget->sig.set(this, default_callback);
}

void Float_Slider::default_valid(bool in)
{
  Float_Base::default_valid(in);
  if (in) _default_widget->show();
  else _default_widget->hide();
  update();
}

void Float_Slider::range(double min, double max)
{
  Float_Base::range(min, max);
  slider_range(min, max);
}

void Float_Slider::slider_range(double min, double max)
{
  _slider->range(min, max);
}

void Float_Slider::label(widget::Widget * in)
{
  if (_label) _layout->del(_label);
  _label = in;
  if (_label) _layout->insert(_label, 0);
  _layout->stretch(_slider);
  update();
}

int Float_Slider::handle(int in)
{
  if (Float_Base::handle(in)) return 1;
  return widget::Widget::handle(in);
}

layout::Item * Float_Slider::group_align() { return _widget->group_align(); }

void Float_Slider::widget_callback(const Str & in)
{
  Float_Base::set(str::str_to_float(in));
  value_update();
}

void Float_Slider::inc_callback(bool)
{
  inc_small_action();
}

void Float_Slider::dec_callback(bool)
{
  dec_small_action();
}

void Float_Slider::slider_callback(double in)
{
  Float_Base::set(in);
}

void Float_Slider::default_callback(bool)
{
  Float_Base::set(default_value());
}

void Float_Slider::value_update()
{
  callbacks(false);
  
  const double value = get();
  
  _widget->set(str::float_to_str(value));
  _slider->set(value);
  
  if (default_valid() && value != default_value()) _default_widget->activate();
  else _default_widget->deactivate();

  callbacks(true);
}

void Float_Slider::range_update()
{
  callbacks(false);
  
  _slider->range(min(), max());
  _slider->inc(inc_small(), inc_large());

  callbacks(true);
}

}
}
}
