/*------------------------------------------------------------------------------
 lib/djv_ui/prefs.cxx
 
 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#include "prefs.h"
#include <djv_image/io.h>
#include <djv_ui/app.h>
#include <FL/Fl_Preferences.H>
#include <stdlib.h>

namespace djv {
namespace ui {
namespace prefs {

Prefs::Prefs(const Str & name) :
  _prefs(new Fl_Preferences(
    Fl_Preferences::USER,
    Str(str::Format("djv-%1").arg(DJV_VERSION)).c_str(),
    name.c_str()))
{}

Prefs::Prefs(const Prefs & in, const Str & group) :
  _prefs(new Fl_Preferences(in._prefs.get(), group.c_str()))
{}

Prefs::Prefs(Prefs * in, const Str & group) :
  _prefs(new Fl_Preferences(in->_prefs.get(), group.c_str()))
{}

Prefs::~Prefs() { _prefs->flush(); }

void Prefs::set(const Str & in, const List<Str> & value)
{
  _prefs->set(in.c_str(), str::join(value, "|").c_str());
}

bool Prefs::get(const Str & in, List<Str> * out) const
{
  char * buf = 0;
  const bool found = _prefs->get(in.c_str(), buf, "") ? true : false;
  const List<Str> tmp = str::split(buf, "|", true);
  if (! (1 == tmp.size() && ! tmp[0].size()))
    *out = tmp;
  free(buf);
  return found;
}
  
void Prefs::del(const Str & in)
{
  _prefs->deleteEntry(in.c_str());
}

void Prefs::del_group(const Str & in)
{
  _prefs->deleteGroup(in.c_str());
}

List<Str> Prefs::list() const
{
  List<Str> out;
  const int size = _prefs->entries();
  for (int i = 0; i < size; ++i)
    out += _prefs->entry(i);
  return out;
}

List<Str> Prefs::group_list() const
{
  List<Str> out;
  const int size = _prefs->groups();
  for (int i = 0; i < size; ++i)
    out += _prefs->group(i);
  return out;
}

namespace {

void prefs_load(const prefs::Prefs & in, djv::image::io::Plugin * plugin)
{
  prefs::Prefs prefs(in, plugin->name());
  const List<Str> & option = plugin->option_list();
  for (uint i = 0; i < option.size(); ++i) {
    List<Str> tmp;
    if (prefs.get(option[i], &tmp)) plugin->option(option[i], &tmp);
  }
}

void prefs_save(prefs::Prefs * in, const djv::image::io::Plugin * plugin)
{
  prefs::Prefs prefs(in, plugin->name());
  const List<Str> & option = plugin->option_list();
  for (uint i = 0; i < option.size(); ++i)
    prefs.set(option[i], plugin->option(option[i]));
}

}

Image::Image()
{
  ui::prefs::Prefs prefs(ui::prefs::global(), "image");
  List<Str> tmp;
  
  if (prefs.get("speed", &tmp)) {
    djv::image::time::Speed value;
    djv::image::time::serialize(&tmp, &value);
    djv::image::time::speed(value);
  }
  
  prefs::Prefs load(prefs, "load");
  const List<djv::image::io::Load *> & load_list =
    djv::image::io::load()->list();
  for (uint i = 0; i < load_list.size(); ++i)
    prefs_load(load, load_list[i]);
  
  prefs::Prefs save(prefs, "save");
  const List<djv::image::io::Save *> & save_list =
    djv::image::io::save()->list();
  for (uint i = 0; i < save_list.size(); ++i)
    prefs_load(save, save_list[i]);
}

Image::~Image()
{
  ui::prefs::Prefs prefs(ui::prefs::global(), "image");
  
  prefs.set("speed", djv::image::time::serialize(
    djv::image::time::speed()));

  prefs::Prefs load(prefs, "load");
  const List<djv::image::io::Load *> & load_list =
    djv::image::io::load()->list();
  for (uint i = 0; i < load_list.size(); ++i)
    prefs_save(&load, load_list[i]);
  
  prefs::Prefs save(prefs, "save");
  const List<djv::image::io::Save *> & save_list =
    djv::image::io::save()->list();
  for (uint i = 0; i < save_list.size(); ++i)
    prefs_save(&save, save_list[i]);
}

void Image::speed(const djv::image::time::Speed & in)
  { djv::image::time::speed(in); }
const djv::image::time::Speed & Image::speed() const
  { return djv::image::time::speed(); }

namespace {
std::auto_ptr<Prefs> _app, _global;
std::auto_ptr<Image> _image;
}

Prefs * app() { return _app.get(); }
Prefs * global() { return _global.get(); }
Image * image() { return _image.get(); }

void init()
{
  _app = std::auto_ptr<Prefs>(new Prefs(app::app()->name()));
  _global = std::auto_ptr<Prefs>(new Prefs("global"));
  _image = std::auto_ptr<Image>(new Image);
}

void del()
{
  _image.reset();
  _app.reset();
  _global.reset();
}

}
}
}
