/*------------------------------------------------------------------------------
 lib/djv_ui/style.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_ui/style.cxx"

#include "style.h"
#include <djv_ui/prefs.h>
#include <djv_ui/app.h>
#include <FL/Fl.H>
#include <FL/Fl_Tooltip.H>

namespace djv {
namespace ui {
namespace style {
using djv::image::Color;

namespace {

const Color _color_foreground [] = { Color(1.0), Color(1.0), Color(0.0) };
const Color _color_background [] = { Color(0.4), Color(0.6), Color(0.8) };
const Color _color_background2 [] = { Color(0.3), Color(0.5), Color(0.7) };
const Color _color_select [] =
  { Color(1.0, 0.9, 0.5), Color(1.0, 0.9, 0.5), Color(1.0, 0.9, 0.5) };

const int _pick_tolerance [] = { 5, 5, 5 };
//const int _handle_size [] = { 20, 40, 60 };

const int _font_size [] = { 10, 12, 14 };

const int _spacing [] = { 4, 4, 4 };
const int _spacing_large [] = { 16, 16, 16 };
const int _margin [] = { 4, 4, 4 };
const int _margin_widget [] = { 4, 4, 4 };
const int _margin_large [] = { 16, 16, 16 };

const int _icon_size = 22;
const int _tool_icon_size = 11;
const int _button_size [] = { 40, 50, 60 };
const int _text_size [] = { 100, 120, 140 };
const int _swatch_size [] = { 40, 50, 60 };
const int _thumbnail_size [] = { 60, 80, 100 };

}

class Style::_Style {
public:

  _Style();
  ~_Style();
  
  void color(COLOR);
  COLOR color() const { return _color; }
  
  void size(SIZE);
  SIZE size() const { return _size; }

  void timecode(bool);
  bool timecode() const { return _timecode; }

  void tooltips(bool);
  bool tooltips() const { return _tooltips; }
  
private:

  void color_update();
  void size_update();
  void timecode_update();
  void tooltips_update();
  
  COLOR _color;
  SIZE _size;
  int _pick_tolerance, _handle_size;
  bool _timecode;
  bool _tooltips;
};

Style::_Style::_Style() :
  _color(COLOR_NEUTRAL),
  _size(SIZE_MEDIUM), _pick_tolerance(5), _handle_size(20),
  _timecode(true),
  _tooltips(true)
{
  DJV_DBG_OBJ2(_DBG, "Style::_Style::_Style");

  prefs::Prefs prefs(prefs::global(), "style");
  List<Str> tmp;
  if (prefs.get("color", &tmp))
    str::serialize(&tmp, &_color, style::color_label());
  if (prefs.get("size", &tmp))
    str::serialize(&tmp, &_size, style::size_label());
  if (prefs.get("timecode", &tmp))
    str::serialize(&tmp, &_timecode);
  if (prefs.get("tooltips", &tmp))
    str::serialize(&tmp, &_tooltips);

  color_update();
  size_update();
  timecode_update();
  tooltips_update();
}

Style::_Style::~_Style()
{
  DJV_DBG_OBJ2(_DBG, "Style::_Style::~_Style");

  prefs::Prefs prefs(prefs::global(), "style");
  prefs.set("color", str::serialize(_color, color_label()));
  prefs.set("size", str::serialize(_size, size_label()));
  prefs.set("timecode", str::serialize(_timecode));
  prefs.set("tooltips", str::serialize(_tooltips));
}

void Style::_Style::color(COLOR in)
{
  if (in == _color) return;
  _color = in;
  color_update();
}

void Style::_Style::color_update()
{
  using namespace djv::image;
  using namespace djv::image::pixel;
  using namespace djv::image::color;

  DJV_DBG_OBJ2(_DBG, "Style::_Style::color_update");

  Color tmp(RGB_U8);
  convert(_color_foreground[_color], &tmp);
  Fl::foreground(tmp.get_u8(0), tmp.get_u8(1), tmp.get_u8(2));
  convert(_color_background[_color], &tmp);
  Fl::background(tmp.get_u8(0), tmp.get_u8(1), tmp.get_u8(2));
  convert(_color_background2[_color], &tmp);
  Fl::background2(tmp.get_u8(0), tmp.get_u8(1), tmp.get_u8(2));
  convert(_color_select[_color], &tmp);
  Fl::set_color(FL_SELECTION_COLOR, tmp.get_u8(0), tmp.get_u8(1),
    tmp.get_u8(2));
  
  app::app()->window_redraw();
}

void Style::_Style::size(SIZE in)
{
  if (in == _size) return;
  _size = in;
  size_update();
}

void Style::_Style::size_update()
{
  DJV_DBG_OBJ2(_DBG, "Style::_Style::size_update");

  Fl_Tooltip::size(_font_size[_size]);
  app::app()->window_update();
}

void Style::_Style::timecode(bool in)
{
  if (in == _timecode) return;
  _timecode = in;
  timecode_update();
}

void Style::_Style::timecode_update()
{
  app::app()->window_update();
}

void Style::_Style::tooltips(bool in)
{
  if (in == _tooltips) return;
  _tooltips = in;
  tooltips_update();
}

void Style::_Style::tooltips_update()
{
  if (_tooltips) Fl_Tooltip::enable();
  else Fl_Tooltip::disable();
}

Style::Style() : _p(new _Style) {}

void Style::color(COLOR in) { _p->color(in); }
COLOR Style::color() const { return _p->color(); }

const Color & Style::color_foreground() const
  { return _color_foreground[_p->color()]; }
const Color & Style::color_background() const
  { return _color_background[_p->color()]; }
const Color & Style::color_background2() const
  { return _color_background2[_p->color()]; }
const Color & Style::color_select() const
  { return _color_select[_p->color()]; }

void Style::size(SIZE in) { _p->size(in); }
SIZE Style::size() const { return _p->size(); }

int Style::pick_tolerance() const { return _pick_tolerance[_p->size()]; }
int Style::handle_size() const { return _pick_tolerance[_p->size()]; }

Fl_Font Style::font() const { return FL_HELVETICA; }
Fl_Font Style::font_bold() const { return FL_HELVETICA_BOLD; }
int Style::font_size() const { return _font_size[_p->size()]; }

void Style::timecode(bool in) { _p->timecode(in); }
bool Style::timecode() const { return _p->timecode(); }

void Style::tooltips(bool in) { _p->tooltips(in); }
bool Style::tooltips() const { return _p->tooltips(); }

int Style::spacing() const { return _spacing[_p->size()]; }
int Style::spacing_large() const { return _spacing_large[_p->size()]; }
int Style::margin() const { return _margin[_p->size()]; }
int Style::margin_widget() const { return _margin_widget[_p->size()]; }
int Style::margin_large() const { return _margin_large[_p->size()]; }

int Style::icon_size() const { return _icon_size; }
int Style::tool_icon_size() const { return _tool_icon_size; }
int Style::button_size() const { return _button_size[_p->size()]; }
int Style::text_size() const { return _text_size[_p->size()]; }
int Style::swatch_size() const { return _swatch_size[_p->size()]; }
int Style::thumbnail_size() const { return _thumbnail_size[_p->size()]; }

const List<Str> & color_label()
{
  static const List<Str> data = List<Str>() << "Dark" << "Neutral" << "Light";
  return data;
}

const List<Str> & size_label()
{
  static const List<Str> data = List<Str>() << "Small" << "Medium" << "Large";
  return data;
}

namespace { std::auto_ptr<Style> _style; }

Style * style()
{
  if (! _style.get()) _style = std::auto_ptr<Style>(new Style);
  return _style.get();
}

void del()
{
  _style.reset();
}

}
}
}
