/*------------------------------------------------------------------------------
 lib/djv_ui/window.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_ui/window.cxx"

#include "window.h"
#include <djv_ui/app.h>
#include <FL/Fl.H>

namespace djv {
namespace ui {
namespace window {

Widget::Widget(const Str & title) :
  Fl_Double_Window(0, 0, 0, 0),
  _title(title),
  _fullscreen(false),
  _size_hint_init(false),
  _resizable(true)
{
  DJV_DBG_OBJ2(_DBG, str::Format("Widget::Widget(%1)").arg(int(this)));

  end();
  
  label(_title.c_str());
  
  callback(close_callback, this);

  app::app()->window_add(this);
}

Widget::~Widget()
{
  DJV_DBG_OBJ2(_DBG, str::Format("Widget::~Widget(%1)").arg(int(this)));
}

void Widget::title(const Str & in)
{
  _title = in;
  
  label(_title.c_str());
}

void Widget::fullscreen(bool in)
{
  if (in == _fullscreen) return;

  DJV_DBG_OBJ2(_DBG, "Widget::fullscreen");
  DJV_DBG_PRN("in = " << in);

  _fullscreen = in;
  
  if (_fullscreen) {
    _geom_tmp = Box2i(Vec2i(x(), y()), geom().size);
    Fl_Double_Window::fullscreen();
  } else {
    fullscreen_off(_geom_tmp.x, _geom_tmp.y, _geom_tmp.w, _geom_tmp.h);
    if (_resizable) {
      const Vec2i min = size_min();
      size_range(min.x, min.y, 0, 0);
    } else {
      size_range(_geom_tmp.w, _geom_tmp.h, _geom_tmp.w, _geom_tmp.h);
    }
    free_position();
  }
}

int Widget::handle(int in)
{
  //DJV_DBG_OBJ2(_DBG, "Widget::handle");
  //DJV_DBG_PRN("in = " << in);
  
  switch (widget::Base::handle(in)) {
    case -1: return 0;
    case 1: return 1;
  }
  //switch (in) {
  //  case FL_PUSH: Fl::focus(this); break;
  //}
  return Fl_Double_Window::handle(in);
}

void Widget::draw()
{
  Fl_Double_Window::draw();
  widget::Base::draw();
}

void Widget::widget_del()
{
  DJV_DBG_OBJ2(_DBG, "Widget::widget_del");

  app::app()->window_del(this);
  widget::Base::widget_del();
}

void Widget::update()
{
  DJV_DBG_OBJ2(_DBG, str::Format("Widget::update(%1)").arg(int(this)));
  DJV_DBG_PRN("resizable = " << _resizable);

  dirty();
  
  const Vec2i min = size_min();
  const Vec2i size = clamp(geom().size, min, Vec2i(Fl::w(), Fl::h()));
  DJV_DBG_PRN("min = " << min);
  DJV_DBG_PRN("size = " << size);

  Fl_Double_Window::size(size.x, size.y);
  if (_resizable) {
    size_range(min.x, min.y, 0, 0);
  } else {
    size_range(size.x, size.y, size.x, size.y);
  }
  widget::Base::resize(0, 0, size.x, size.y);
  redraw();
}

void Widget::size(const Vec2i & in, bool resizable)
{
  DJV_DBG_OBJ2(_DBG, str::Format("Widget::size(%1)").arg(int(this)));
  DJV_DBG_PRN("in = " << in);
  DJV_DBG_PRN("resizable = " << resizable);
  
  const Vec2i min = size_min();
  const Vec2i size = clamp(in, min, Vec2i(Fl::w(), Fl::h()));
  DJV_DBG_PRN("min = " << min);
  DJV_DBG_PRN("size = " << size);
  
  if (_fullscreen) {
    fullscreen_off(_geom_tmp.x, _geom_tmp.y, size.x, size.y);
    _fullscreen = false;
  } else {
    Fl_Double_Window::size(size.x, size.y);
  }

  _resizable = resizable;
  if (_resizable) {
    size_range(min.x, min.y, 0, 0);
  } else {
    size_range(size.x, size.y, size.x, size.y);
  }

  free_position();
  update();
}

void Widget::resize(int x, int y, int w, int h)
{
  DJV_DBG_OBJ2(_DBG, str::Format("Widget::resize(%1)").arg(int(this)));
  DJV_DBG_PRN("in = " << Box2i(x, y, w, h));

  const bool dirty = Vec2i(w, h) != geom().size;
  DJV_DBG_PRN("dirty = " << dirty);
  
  Fl_Double_Window::resize(x, y, w, h);
  
  if (dirty) {
    widget::Base::resize(0, 0, w, h);
    //update();
    //if (layout()) layout()->dbg();
  }
}

Vec2i Widget::size_hint() const
{
  // XXX
  
  if (! _size_hint_init) {
    Widget * that = const_cast<Widget *>(this);
    that->_size_hint_init = true;
    that->update();
  }
  
  return widget::Base::size_hint();
}

Vec2i Widget::size_min() const
{
  //return Vec2i(100, 100);
  return max(widget::Base::size_min(), size_hint());
}

void Widget::close_callback()
{
  hide();
  close_sig.emit(true);
}

}
}
}
