/*------------------------------------------------------------------------------
 plugin/djv_jpeg/save.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "plugin/djv_jpeg/save.cxx"

#include "plugin.h"

using namespace image;
using libjpeg::jpeg_compress_struct;

extern "C" {
DJV_PLUGIN_EXPORT plugin::Plugin * djv_image_save() { return new Save; }
}

Save::Save() : _f(0), _jpeg_init(false) {}
Save::~Save() { _close(); }

Str Save::help() const
{
  return str::Format(
"%1:\n"
"* quality (value) - Compression quality: 1-100. Default: %2.\n"
  ).
    arg(plugin_name).
    arg(_options.quality);
}

plugin::Plugin * Save::copy() const
  { Save * out = new Save; out->_options = _options; return out; }

bool Save::option(const Str & in, List<Str> * data)
{
  if (str::compare(in, option_list()[OPTION_QUALITY]))
    return str::serialize(data, &_options.quality);
  return false;
}

List<Str> Save::option(const Str & in) const
{
  if (str::compare(in, option_list()[OPTION_QUALITY]))
    return str::serialize(_options.quality);
  return List<Str>();
}

const List<Str> & Save::option_list() const { return save_option_label(); }

void Save::open(const File & in, const image::io::Info & info) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::open");
  DJV_DBG_PRN("in = " << in);
  
  _file = in;
  
  // Temporary buffer.
  
  Info _info;
  _info.size = info.size;
  
  pixel::FORMAT format = pixel::format(info.pixel);
  switch (format) {
    case pixel::LA: format = pixel::L; break;
    case pixel::RGBA: format = pixel::RGB; break;
    default: break;
  }
  _info.pixel = pixel::pixel(format, pixel::U8);
  
  _image.set(_info);
  DJV_DBG_PRN("image = " << _image);
}

namespace {

bool _jpeg_init(jpeg_compress_struct * jpeg, Jpeg_Error * error)
{
  if (setjmp(error->jump)) return false;
  libjpeg::jpeg_create_compress(jpeg);
  return true;
}

bool _jpeg_open(FILE * f, jpeg_compress_struct * jpeg,
  const image::Info & info, int quality, Jpeg_Error * error
) {
  if (setjmp(error->jump)) return false;

  libjpeg::jpeg_stdio_dest(jpeg, f);
  
  jpeg->image_width = info.size.x;
  jpeg->image_height = info.size.y;
  
  if (pixel::L_U8 == info.pixel) {
    jpeg->input_components = 1;
    jpeg->in_color_space = libjpeg::JCS_GRAYSCALE;
  } else if (pixel::RGB_U8 == info.pixel) {
    jpeg->input_components = 3;
    jpeg->in_color_space = libjpeg::JCS_RGB;
  }
  
  libjpeg::jpeg_set_defaults(jpeg);
  libjpeg::jpeg_set_quality(jpeg, quality, TRUE);
  libjpeg::jpeg_start_compress(jpeg, TRUE);
  
  Str tag = info.tag[tag::label()[tag::DESCRIPTION]];
  if (tag.size())
    libjpeg::jpeg_write_marker(
      jpeg, JPEG_COM, (libjpeg::JOCTET *)tag.c_str(), tag.size());
  
  return true;
}

}

void Save::_open(const File & in, const tag::Tag & tag) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::_open");
  DJV_DBG_PRN("in = " << in);
  
  _close();
  
  // Initialize libjpeg.
  
  _jpeg.err = libjpeg::jpeg_std_error(&_jpeg_error.pub);
  _jpeg_error.pub.error_exit = djv_jpeg_error;
  _jpeg_error.pub.emit_message = djv_jpeg_warn;
  _jpeg_error.msg[0] = 0;
  
  if (! ::_jpeg_init(&_jpeg, &_jpeg_error))
    throw Error(plugin_name, _jpeg_error.msg);
  
  _jpeg_init = true;
  
  // Open.
  
  _f = ::fopen(in.get().c_str(), "wb");
  if (! _f)
    throw Error(plugin_name, image::io::error_open);
  
  image::Info info = _image.info();
  info.tag = tag;
  if (! _jpeg_open(_f, &_jpeg, info, _options.quality, &_jpeg_error))
    throw Error(plugin_name, _jpeg_error.msg);
}

void Save::_close()
{
  if (_jpeg_init) {
    libjpeg::jpeg_destroy_compress(&_jpeg);
    _jpeg_init = false;
  }
  if (_f) { ::fclose(_f); _f = 0; }
}

namespace {

bool _scanline(libjpeg::jpeg_compress_struct * jpeg, const uint8_t * in,
  Jpeg_Error * error)
{
  if (::setjmp(error->jump)) return false;
  libjpeg::JSAMPROW p [] = { (libjpeg::JSAMPLE *)in };
  if (! libjpeg::jpeg_write_scanlines(jpeg, p, 1)) return false;
  return true;
}

bool _end(libjpeg::jpeg_compress_struct * jpeg, Jpeg_Error * error)
{
  if (::setjmp(error->jump)) return false;
  libjpeg::jpeg_finish_compress(jpeg);
  return true;
}

}

void Save::set(const Image & in, int frame) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::set");
  DJV_DBG_PRN("in = " << in);
  
  // Open.

  _open(_file.get(frame), in.info().tag);
  
  // Temporary buffers.
  
  const Image * image_p = &in;
  if (in.info() != _image.info()) {
    DJV_DBG_PRN("convert = " << _image);
    _image.zero();
    gl::data::copy(in, &_image);
    image_p = &_image;
  }
  
  // Write.
  
  const int h = image_p->h();
  for (int y = 0; y < h; ++y)
    if (! _scanline(&_jpeg, image_p->data(0, h - 1 - y), &_jpeg_error))
      throw Error(plugin_name, _jpeg_error.msg);
  
  if (! _end(&_jpeg, &_jpeg_error))
    throw Error(plugin_name, _jpeg_error.msg);

  _close();
}

